/*
 * DetectionCheckTable.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.database;

import java.io.*;
import java.util.*;
import misao.pixy.*;
import misao.util.*;
import misao.host.*;

public class DetectionCheckTable {
    private CelestialMapIndex map;

    private double d_pixel;
    private double d_mag;

    private Vector ignore_image_list;

    String[] pxf_path;

    PrintManager print_manager;

    public DetectionCheckTable ( double _d_pixel, double _d_mag ) {
	map = new CelestialMapIndex();

	d_pixel = _d_pixel;
	d_mag = _d_mag;

	ignore_image_list = new Vector();

	pxf_path = null;

	print_manager = null;
    }

    public void setPrintManager ( PrintManager _print_manager ) {
	print_manager = _print_manager;
    }

    public void setIgnoreImage ( Vector l ) {
	for (int i = 0 ; i < l.size() ; i++)
	    ignore_image_list.addElement(l.elementAt(i));
    }

    public boolean isIgnored ( String image ) {
	for (int i = 0 ; i < ignore_image_list.size() ; i++) {
	    ImageInformation info = (ImageInformation)ignore_image_list.elementAt(i);
	    if (info.image_url.equals(image))
		return true;
	}
	return false;
    }

    public void setPxfPath ( String[] _pxf_path ) {
	pxf_path = _pxf_path;
    }

    public void add ( PairData pair, ImageInformation image_info ) throws Exception {
	double pixel_size = image_info.field_of_view / (double)image_info.image_size.getWidth() * 3600.0;

	StarData sd = pair.getFirstStarData();
	MagnitudeDatabaseStar mag_star = findStar(sd.getCoor(), pixel_size);
	if (mag_star == null) {
	    double pos_error = (image_info.pos_error.getRA() + image_info.pos_error.getDecl()) / 2.0 * 3.0;
	    if (pos_error < 5.0)
		pos_error = 5.0;
	    MisaoPreliminaryObject ref_name = new MisaoPreliminaryObject(sd.getCoor(), pos_error);
	    mag_star = new MagnitudeDatabaseStar(ref_name);
	    map.addElement(mag_star);
	}
	MagnitudeDatabaseRecord record = new MagnitudeDatabaseRecord(pair, image_info);
	mag_star.addRecord(record);
    }

    public MagnitudeDatabaseStar findStar ( Coor coor, double pixel_size ) {
	Vector stars = map.findElementByPosition(coor, pixel_size * d_pixel / 3600.0);

	for (int i = 0 ; i < stars.size() ; i++) {
	    MagnitudeDatabaseStar mag_star = (MagnitudeDatabaseStar)stars.elementAt(i);
	    MisaoPreliminaryObject ref_name = (MisaoPreliminaryObject)mag_star.getReferenceName();
	    double r = ref_name.getCoor().angular_distance(coor) * 3600.0;
	    if (r < pixel_size * d_pixel  ||  r < ref_name.getPositionError())
		return mag_star;
	}

	return null;
    }

    public void removeStarsWithOnlyOneRecord ( ) {
	CelestialMapIndex new_map = new CelestialMapIndex();

	Vector list = map.getAllElements();
	for (int i = 0 ; i < list.size() ; i++) {
	    MagnitudeDatabaseStar mag_star = (MagnitudeDatabaseStar)list.elementAt(i);
	    if (mag_star.countRecord() > 1)
		new_map.addElement(mag_star);
	}

	map = new_map;
    }

    public void checkMagnitudeDatabase ( ) throws Exception {
	MagnitudeDatabaseManager mag_db_manager = new MagnitudeDatabaseManager();
	mag_db_manager.setReadOnly();

	Vector stars = map.getAllElements();

	ProgressManager progress_manager = null;
	if (print_manager != null) {
	    print_manager.println("Reading magnitude database for " + stars.size() + " stars...");
	    progress_manager = new ProgressManager(print_manager);
	    progress_manager.setTotalCount(stars.size());
	    progress_manager.setModePercentByTen();
	}

	for (int i = 0 ; i < stars.size() ; i++) {
	    if (progress_manager != null)
		progress_manager.countUp();

	    MagnitudeDatabaseStar mag_star = (MagnitudeDatabaseStar)stars.elementAt(i);
	    MisaoPreliminaryObject o = (MisaoPreliminaryObject)mag_star.getReferenceName();

	    MagnitudeDatabaseStar star = mag_db_manager.findMisaoObject(o.getCoor(), o.getPositionError());
	    if (star != null) {
		MagnitudeDatabaseRecord record = star.getFirstRecord();
		while (record != null) {
		    if (! isIgnored(record.getImage()))
			mag_star.addRecord(record);
		    record = star.getNextRecord();
		}
	    }
	}
    }

    public void checkPastPxf ( ImageInformation image_info ) {
	if (isIgnored(image_info.image_url))
	    return;

	MapIndex map_pxf = null;

	double dr = image_info.field_of_view / 2.0 * 1.5;
	Vector stars = map.findElementByPosition(image_info.center_coor, dr);

	for (int i = 0 ; i < stars.size() ; i++) {
	    MagnitudeDatabaseStar mag_star = (MagnitudeDatabaseStar)stars.elementAt(i);

	    Position pos = new CoorMapping(mag_star.getPosition().getX(), mag_star.getPosition().getY()).mappingXY(image_info.center_coor);
	    pos.rescale((double)image_info.image_size.getWidth() / image_info.field_of_view);
	    pos = new MapFunction(1.0, 360.0 - image_info.pa, new Position()).inverse().map(pos);
	    Position df = image_info.distortion_field.getValue(pos);
	    df.rescale(-1.0);
	    pos.add(df);
	    pos.add(new Position((double)image_info.image_size.getWidth() / 2.0, (double)image_info.image_size.getHeight() / 2.0));

	    if (0 <= pos.getX()  &&  pos.getX() < image_info.image_size.getWidth()  &&  0 <= pos.getY()  &&  pos.getY() < image_info.image_size.getHeight()) {
		MagnitudeDatabaseRecord record = mag_star.getRecordOfImage(image_info.image_url);
		if (record == null) {
		    record = new MagnitudeDatabaseRecord(null, image_info);
		    record.setPosition(pos);

		    if (map_pxf == null)
			map_pxf = readPxfFile(image_info);

		    Vector l = map_pxf.findElementByPosition(pos, d_pixel, d_pixel);
		    MagnitudeDatabaseStar star = null;
		    double r = 9999.9;
		    for (int j = 0 ; j < l.size() ; j++) {
			MagnitudeDatabaseStar st = (MagnitudeDatabaseStar)l.elementAt(j);
			double r2 = (st.getPosition().getX() - pos.getX()) * (st.getPosition().getX() - pos.getX()) + (st.getPosition().getY() - pos.getY()) * (st.getPosition().getY() - pos.getY());
			if (r2 < r  &&  r2 < d_pixel * d_pixel) {
			    r = r2;
			    star = st;
			}
		    }
		    if (star != null)
			record = new MagnitudeDatabaseRecord(star.getFirstRecord().getRecord());
		}

		mag_star.addRecord(record);
	    }
	}
    }

    private MapIndex readPxfFile ( ImageInformation image_info ) {
	MapIndex map_pxf = new MapIndex(0, 0, image_info.image_size.getWidth(), image_info.image_size.getHeight(), 30, 30);

	File file = findPxfFile(image_info, null);
	if (file == null  &&  pxf_path != null) {
	    for (int i = 0 ; i < pxf_path.length ; i++) {
		file = findPxfFile(image_info, pxf_path[i]);
		if (file != null)
		    break;
	    }
	}
	if (file != null) {
	    try {
		FileInputStream fis = new FileInputStream(file);
		DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
		PxfFile pxf = new PxfFile();
		pxf.read(dis);
		dis.close();

		Vector list = pxf.pair_list.getAllList();
		for (int i = 0 ; i < list.size() ; i++) {
		    Vector l = (Vector)list.elementAt(i);

		    for (int j = 0 ; j < l.size() ; j++) {
			PairData pair = (PairData)l.elementAt(j);
			StarData sdd = (StarData)pair.getFirstData();
			StarData sdc = (StarData)pair.getSecondData();
			StarData sd = sdc;
			if (sd == null)
			    sd = sdd;
			Position p = sd.getPosition();
			if (sdd != null)
			    p = sdd.getPosition();

			MagnitudeDatabaseRecord record = new MagnitudeDatabaseRecord(pair, image_info);
			MagnitudeDatabaseStar mag_star = new MagnitudeDatabaseStar(new MisaoPreliminaryObject(new Coor(p.getX(), p.getY()), 0));
			mag_star.addRecord(record);
			map_pxf.addElement(mag_star);
		    }
		}
	    } catch (Exception e) {
	    }
	}

	return map_pxf;
    }

    private static File findPxfFile ( ImageInformation image_info, String path ) {
	String image_url = image_info.image_url;
	if (path != null)
	    image_url = path + "/" + image_url;

	File file = new File(image_url + ".pxf");
	if (file.canRead())
	    return file;

	file = new File(image_url + ".PXF");
	if (file.canRead())
	    return file;

	String pxf_filename = image_url;
	int p = pxf_filename.lastIndexOf('.');
	if (p >= 0) {
	    pxf_filename = pxf_filename.substring(0,p);

	    file = new File(pxf_filename + ".pxf");
	    if (file.canRead())
		return file;

	    file = new File(pxf_filename + ".PXF");
	    if (file.canRead())
		return file;
	}

	return null;
    }

    private static int MODE_NEW_OBJECT = 1;
    private static int MODE_VARIABLE = 2;
    private static int MODE_NORMAL = 3;
    private static int MODE_NO_PAST = 4;

    public Vector getNewObjectList ( ) {
	return getList(MODE_NEW_OBJECT);
    }
    public Vector getVariableList ( ) {
	return getList(MODE_VARIABLE);
    }
    public Vector getNormalList ( ) {
	return getList(MODE_NORMAL);
    }
    public Vector getNoPastList ( ) {
	return getList(MODE_NO_PAST);
    }
    private Vector getList ( int mode ) {
	Vector list = new Vector();

	Vector stars = map.getAllElements();
	for (int i = 0 ; i < stars.size() ; i++) {
	    MagnitudeDatabaseStar mag_star = (MagnitudeDatabaseStar)stars.elementAt(i);

	    double recent_mag = 0.0;
	    int recent_count = 0;
	    double brightest_in_past = 999.9;
	    double faintest_in_past = -999.9;
	    int detected_in_past = 0;
	    double faintest_upper_limit = -999.9;

	    MagnitudeDatabaseRecord record = mag_star.getFirstRecord();
	    while (record != null) {
		double mag = record.getMag();
		boolean visible_flag = record.isDetected();

		if (isIgnored(record.getImage())) {
		    if (visible_flag) {
			recent_mag += mag;
			recent_count++;
		    }
		} else {
		    if (visible_flag) {
			if (brightest_in_past > mag)
			    brightest_in_past = mag;
			if (faintest_in_past < mag)
			    faintest_in_past = mag;
			detected_in_past++;
		    } else {
			if (faintest_upper_limit < mag)
			    faintest_upper_limit = mag;
		    }
		}

		record = mag_star.getNextRecord();
	    }

	    recent_mag /= (double)recent_count;

	    if (detected_in_past == 0) {
		if (faintest_upper_limit >= recent_mag + 1.0) {
		    if (mode == MODE_NEW_OBJECT)
			list.addElement(mag_star);
		} else {
		    if (mode == MODE_NO_PAST)
			list.addElement(mag_star);
		}
	    } else if (Math.abs(recent_mag - brightest_in_past) < 1.0  &&  Math.abs(recent_mag - faintest_in_past) < 1.0  &&  faintest_upper_limit < recent_mag + 1.0) {
		if (mode == MODE_NORMAL)
		    list.addElement(mag_star);
	    } else {
		if (mode == MODE_VARIABLE)
		    list.addElement(mag_star);
	    }
	}

	return list;
    }
}
