/*
 * MagnitudeDatabaseManager.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.database;

import java.io.*;
import java.util.*;
import misao.util.*;
import misao.pixy.*;
import misao.host.*;

public class MagnitudeDatabaseManager {
    private boolean read_only;
    private boolean to_delete;

    private MagnitudeDatabaseStar mag_star;
    private String mag_star_filename;

    private CelestialMapIndex map_index_iras_to_be_updated;
    private MapIndex map_index_iras;
    private String iras_filename;

    private CelestialMapIndex map_index_misao_to_be_updated;
    private MapIndex map_index_misao;
    private String misao_filename;

    // temporarily used in findMisaoObjectCore()
    private String found_misao_filename;
    private String found_misao_dir;
    private MagnitudeDatabaseStar found_misao_star;
    private double found_misao_distance;

    public MagnitudeDatabaseManager ( ) {
	read_only = false;
	to_delete = false;

	mag_star = null;
	mag_star_filename = "";

	map_index_iras_to_be_updated = new CelestialMapIndex();
	map_index_iras = null;
	iras_filename = "";

	map_index_misao_to_be_updated = new CelestialMapIndex();
	map_index_misao = null;
	misao_filename = "";
    }

    public void setReadOnly ( ) {
	read_only = true;
    }

    public void setToDelete ( ) {
	to_delete = true;
    }

    public void close ( ) throws Exception {
	if (read_only == false) {
	    closeStar();
	    updateIras();
	    updateMisao();
	}
    }

    private void openStar ( ) throws Exception {
	if (mag_star != null) {
	    File file = new File(mag_star_filename);
	    if (file.exists()) {
		FileInputStream fis = new FileInputStream(file);
		DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
		mag_star.read(dis);
		dis.close();
	    }
	}
    }

    private void closeStar ( ) throws Exception {
	if (read_only == false) {
	    if (mag_star != null) {
		FileOutputStream os = new FileOutputStream(mag_star_filename);
		PrintStream ps = new PrintStream(os);
		mag_star.print(ps);
		ps.close();
	    }

	    mag_star = null;
	    mag_star_filename = "";
	}
    }

    private void openIras ( ) throws Exception {
	File file = new File(iras_filename);
	if (file.exists()) {
	    FileInputStream fis = new FileInputStream(file);
	    DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));

	    MagnitudeDatabaseStar star = null;
	    String record = dis.readLine();
	    while (record != null) {
		if (record.substring(0,4).equals("IRAS")) {
		    ReferenceName ref_name = new ReferenceName(record, ReferenceType.REF_S);
		    star = new MagnitudeDatabaseStar(ref_name);
		    map_index_iras.addElement(star);
		} else {
		    MagnitudeDatabaseRecord mag_record = new MagnitudeDatabaseRecord(record);
		    star.addRecord(mag_record);
		}

		record = dis.readLine();
	    }

	    dis.close();
	}
    }

    private void closeIras ( ) {
	if (read_only == false) {
	    try {
		FileOutputStream os = new FileOutputStream(iras_filename);
		PrintStream ps = new PrintStream(os);

		Vector list = map_index_iras.getAllElements();
		for (int i = 0 ; i < list.size() ; i++) {
		    MagnitudeDatabaseStar star = (MagnitudeDatabaseStar)list.elementAt(i);
		    if (star.empty() == false) {
			ps.println(star.getReferenceName().getName());
			star.print(ps);
		    }
		}

		ps.close();
	    } catch ( Exception e ) {
	    }
	}

	map_index_iras_to_be_updated = new CelestialMapIndex();
	map_index_iras = null;
	iras_filename = "";
    }

    private void updateIras ( ) throws Exception {
	if (read_only == false) {
	    Vector list = map_index_iras_to_be_updated.getAllElements();

	    for (int i = 0 ; i < list.size() ; i++) {
		MagnitudeDatabaseUpdateElement elm = (MagnitudeDatabaseUpdateElement)list.elementAt(i);
		VariableStarName vsname = new VariableStarName(elm.ref_name.getName());
		String filename = MisaoProperty.getObsPath() + "/suspicious/" + vsname.getDir();

		if (! filename.equals(iras_filename)) {
		    if (iras_filename.length() > 0)
			closeIras();

		    Coor coor = IrasObject.getCoor(elm.ref_name);
		    Position pos = new Position(coor.getRA(), coor.getDecl());
		    int ra_base = map_index_iras_to_be_updated.getFirstIndex(pos) * 15;
		    int decl_base = map_index_iras_to_be_updated.getSecondIndex(pos) * 10 - 90;
		    map_index_iras = new MapIndex(ra_base, ra_base + 15, decl_base, decl_base + 10, 30, 20);
		    iras_filename = filename;
		    openIras();
		}

		MagnitudeDatabaseStar star = (MagnitudeDatabaseStar)map_index_iras.findElementByName(elm.getName(), elm.getPosition());
		if (star == null) {
		    star = new MagnitudeDatabaseStar(elm.ref_name);
		    map_index_iras.addElement(star);
		}

		if (to_delete)
		    star.removeRecord(elm.mag_record);
		else
		    star.addRecord(elm.mag_record);
	    }

	    closeIras();
	}
    }

    private void openMisao ( ) throws Exception {
	File file = new File(misao_filename);
	if (file.exists()) {
	    FileInputStream fis = new FileInputStream(file);
	    DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));

	    MagnitudeDatabaseStar star = null;
	    String record = dis.readLine();
	    while (record != null) {
		if (record.charAt(0) == 'J') {
		    MisaoPreliminaryObject ref_name = new MisaoPreliminaryObject(record);
		    star = new MagnitudeDatabaseStar(ref_name);
		    map_index_misao.addElement(star);
		} else {
		    MagnitudeDatabaseRecord mag_record = new MagnitudeDatabaseRecord(record);
		    star.addRecord(mag_record);
		}

		record = dis.readLine();
	    }

	    dis.close();
	}
    }

    private void closeMisao ( ) throws Exception {
	if (read_only == false) {
	    try {
		FileOutputStream os = new FileOutputStream(misao_filename);
		PrintStream ps = new PrintStream(os);

		Vector list = map_index_misao.getAllElements();
		for (int i = 0 ; i < list.size() ; i++) {
		    MagnitudeDatabaseStar star = (MagnitudeDatabaseStar)list.elementAt(i);
		    if (star.empty() == false) {
			ps.println(star.getReferenceName().getName());
			star.print(ps);
		    }
		}

		ps.close();
	    } catch ( Exception e ) {
	    }
	}

	map_index_misao_to_be_updated = new CelestialMapIndex();
	map_index_misao = null;
	misao_filename = "";
    }

    private void findMisaoObjectCore ( MisaoPreliminaryObject o ) throws Exception {
	found_misao_filename = "";
	found_misao_dir = "";
	found_misao_star = null;
	found_misao_distance = 999.9;

	MisaoPreliminaryObjectName vsname = new MisaoPreliminaryObjectName(o.getName());

	Vector dirs = vsname.getAllDirs();
	for (int j = 0 ; j < dirs.size() ; j++) {
	    String dir = (String)dirs.elementAt(j);
	    String filename = MisaoProperty.getObsPath() + "/suspicious/" + dir;

	    if (! filename.equals(misao_filename)) {
		if (misao_filename.length() > 0)
		    closeMisao();

		int ra_base = Integer.valueOf(dir.substring(6,8)).intValue() * 15;
		int decl_base = Integer.valueOf(dir.substring(9,11)).intValue() * 10;
		if (dir.charAt(8) == '-')
		    decl_base = - (decl_base + 10);
		map_index_misao = new MapIndex(ra_base, ra_base + 15, decl_base, decl_base + 10, 30, 20);
		misao_filename = filename;
		openMisao();
	    }

	    double d_ra = 100.0 / 3600.0 / Math.cos(o.getCoor().getDecl() * Astro.RAD);
	    if (Math.abs(o.getCoor().getDecl()) > 89.95)
		d_ra = 180.0;
	    double d_decl = 100.0 / 3600.0;

	    Vector stars = map_index_misao.findElementByPosition(new Position(o.getCoor().getRA(), o.getCoor().getDecl()), d_ra, d_decl);
	    for (int k = 0 ; k < stars.size() ; k++) {
		MagnitudeDatabaseStar star = (MagnitudeDatabaseStar)stars.elementAt(k);
		MisaoPreliminaryObject obj = (MisaoPreliminaryObject)star.getReferenceName();
		double r = obj.getCoor().angular_distance(o.getCoor()) * 3600.0;
		if (r < o.getPositionError()  ||  r < obj.getPositionError()) {
		    if (r < found_misao_distance) {
			found_misao_filename = misao_filename;
			found_misao_dir = dir;
			found_misao_star = star;
			found_misao_distance = r;
		    }
		}
	    }
	}
    }

    public MagnitudeDatabaseStar findMisaoObject ( Coor coor, double d_pos ) throws Exception {
	MisaoPreliminaryObject o = new MisaoPreliminaryObject(coor, d_pos);
	findMisaoObjectCore(o);
	return found_misao_star;
    }

    private void updateMisao ( ) throws Exception {
	if (read_only == false) {
	    Vector list = map_index_misao_to_be_updated.getAllElements();

	    for (int i = 0 ; i < list.size() ; i++) {
		MagnitudeDatabaseUpdateElement elm = (MagnitudeDatabaseUpdateElement)list.elementAt(i);
		MisaoPreliminaryObject o = (MisaoPreliminaryObject)elm.ref_name;
		MisaoPreliminaryObjectName vsname = new MisaoPreliminaryObjectName(o.getName());

		// find the same star in the database
		findMisaoObjectCore(o);
		if (found_misao_filename.length() == 0) {
		    found_misao_filename = MisaoProperty.getObsPath() + "/suspicious/" + vsname.getDir();
		    found_misao_dir = vsname.getDir();
		}

		// open the file again of the same star in the database
		if (! found_misao_filename.equals(misao_filename)) {
		    if (misao_filename.length() > 0)
			closeMisao();

		    int ra_base = Integer.valueOf(found_misao_dir.substring(6,8)).intValue() * 15;
		    int decl_base = Integer.valueOf(found_misao_dir.substring(9,11)).intValue() * 10;
		    if (found_misao_dir.charAt(8) == '-')
			decl_base = - (decl_base + 10);
		    map_index_misao = new MapIndex(ra_base, ra_base + 15, decl_base, decl_base + 10, 30, 20);
		    misao_filename = found_misao_filename;
		    openMisao();
		}

		MagnitudeDatabaseStar star = null;
		if (found_misao_star != null)
		    star = (MagnitudeDatabaseStar)map_index_misao.findElementByName(found_misao_star.getName(), found_misao_star.getPosition());
		if (star == null) {
		    star = new MagnitudeDatabaseStar(elm.ref_name);
		    map_index_misao.addElement(star);
		}

		if (to_delete)
		    star.removeRecord(elm.mag_record);
		else
		    star.addRecord(elm.mag_record);

		// update the name (position and error)
		MisaoPreliminaryObject obj = (MisaoPreliminaryObject)star.getReferenceName();
		if (o.getPositionError() < obj.getPositionError()) {
		    map_index_misao.removeElementByName(star.getName(), star.getPosition());
		    star.setReferenceName(o);

		    // move to another file
		    MisaoPreliminaryObjectName vsn = new MisaoPreliminaryObjectName(obj.getName());
		    if (! vsname.getDir().equals(vsn.getDir())) {
			closeMisao();

			String filename = MisaoProperty.getObsPath() + "/suspicious/" + vsname.getDir();
			int ra_base = Integer.valueOf(vsname.getDir().substring(6,8)).intValue() * 15;
			int decl_base = Integer.valueOf(vsname.getDir().substring(9,11)).intValue() * 10;
			if (vsname.getDir().charAt(8) == '-')
			    decl_base = - (decl_base + 10);
			map_index_misao = new MapIndex(ra_base, ra_base + 15, decl_base, decl_base + 10, 30, 20);
			misao_filename = filename;
			openMisao();
		    }

		    map_index_misao.addElement(star);
		}
	    }

	    closeMisao();
	}
    }

    public void updateRecord ( ReferenceType ref_type, ReferenceName ref_name, MagnitudeDatabaseRecord mag_record ) throws Exception {
	VariableStarName vsname = new VariableStarName(ref_name.getName());
	if (ref_name instanceof MisaoPreliminaryObject)
	    vsname = new MisaoPreliminaryObjectName(ref_name.getName());

	String obs_path = MisaoProperty.getObsPath();
	if (ref_type == ReferenceType.REF_V) {
	    obs_path += "/variable";
	} else if (ref_type == ReferenceType.REF_S) {
	    obs_path += "/suspicious";
	} else if (ref_type == ReferenceType.REF_E) {
	    obs_path += "/error";
	}

	StringTokenizer dir_st = new StringTokenizer(vsname.getDir(), "/");
	String dir_s = obs_path;
	while (dir_st.hasMoreTokens()) {
	    String dir_st_token = dir_st.nextToken();
	    dir_s += "/" + dir_st_token;
	    File directory = new File(dir_s);
	    if (! directory.isDirectory())
		directory.mkdir();

	    if (dir_st_token.equals("iras")  ||  dir_st_token.equals("misao"))
		break;
	}

	if (IrasObject.isIrasObject(ref_name)) {
	    // IRAS object
	    MagnitudeDatabaseUpdateElement elm = new MagnitudeDatabaseUpdateElement(ref_name, mag_record);
	    map_index_iras_to_be_updated.addElement(elm);
	} else if (ref_name instanceof MisaoPreliminaryObject) {
	    // MISAO preliminary object
	    MagnitudeDatabaseUpdateElement elm = new MagnitudeDatabaseUpdateElement(ref_name, mag_record);
	    map_index_misao_to_be_updated.addElement(elm);
	} else {
	    if (mag_star == null  ||  ! mag_star.getReferenceName().getName().equals(ref_name.getName())) {
		if (mag_star != null)
		    closeStar();

		mag_star = new MagnitudeDatabaseStar(ref_name);
		mag_star_filename = obs_path + "/" + vsname.getPath();
		openStar();
	    }

	    if (to_delete)
		mag_star.removeRecord(mag_record);
	    else
		mag_star.addRecord(mag_record);
	}
    }
}
