/*
 * ReferenceCatalogBaseManager.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.host;

import java.io.*;
import java.util.*;
import misao.pixy.*;
import misao.util.*;

public class ReferenceCatalogBaseManager {
    private static final int N = 4;	// must be 4 or more
    private static final int MAX_DATA_SIZE = 32000;

    private ReferenceCatalogBase ref_catalog;
    private String[] ref_file_list;
    private Vector[] ref_data_list;
    private boolean[][] used_flag_list;
    private boolean[] modified_flag_list;
    private int used_flag_phase;

    public ReferenceCatalogBaseManager ( ) {
	ref_catalog = new ReferenceCatalogBase();

	ref_file_list = new String[N];
	ref_data_list = new Vector[N];

	used_flag_list = new boolean[N][N];
	modified_flag_list = new boolean[N];
	used_flag_phase = 0;
	for (int i = 0 ; i < N ; i++) {
	    for (int j = 0 ; j < N ; j++)
		used_flag_list[i][j] = false;
	    modified_flag_list[i] = false;
	}
    }

    private void initPhase ( ) {
	used_flag_phase++;
	if (used_flag_phase == N)
	    used_flag_phase = 0;
	for (int i = 0 ; i < N ; i++)
	    used_flag_list[i][used_flag_phase] = false;
    }

    private int useFile ( String filename ) {
	int p = findIndex(filename);
	if (p < 0) {
	    p = loadableIndex();
	    for (int j = 0 ; j < N ; j++)
		used_flag_list[p][j] = true;
	    load(p, filename);
	}
	used_flag_list[p][used_flag_phase] = true;
	return p;
    }

    public Vector findData ( Coor coor, double r ) {
	Vector list = new Vector();

	initPhase();

	Vector file_list = ref_catalog.getFileList(coor, r);
	for (int i = 0 ; i < file_list.size() ; i++) {
	    String filename = (String)file_list.elementAt(i);
	    int p = useFile(filename);

	    for (int j = 0 ; j < ref_data_list[p].size() ; j++) {
		ReferenceData data = (ReferenceData)ref_data_list[p].elementAt(j);
		if (data.getCoor().angular_distance(coor) < r) {
		    list.addElement(data);
		}
	    }
	}

	return list;
    }

    public ReferenceData findData ( ReferenceData data ) {
	initPhase();

	String filename = ref_catalog.getFile(data.getCoor());
	int p = useFile(filename);

	for (int j = 0 ; j < ref_data_list[p].size() ; j++) {
	    ReferenceData ref_data = (ReferenceData)ref_data_list[p].elementAt(j);
	    if (ref_data.getName().equals(data.getName()))
		return ref_data;
	}

	return null;
    }

    public int findIndex ( String filename ) {
	for (int i = 0 ; i < N ; i++) {
	    if (ref_file_list[i] != null  &&  ref_file_list[i].equals(filename))
		return i;
	}
	return -1;
    }

    public void addData ( ReferenceData data ) {
	replaceData(data, 1);
    }

    public void removeData ( ReferenceData data ) {
	replaceData(data, 2);
    }

    public void replaceData ( ReferenceData data ) {
	replaceData(data, 0);
    }

    private void replaceData ( ReferenceData data, int mode ) {
	initPhase();

	String filename = ref_catalog.getFile(data.getCoor());
	int p = useFile(filename);

	if (mode == 1) {
	    ref_data_list[p].addElement(data);
	} else {
	    boolean replaced_flag = false;
	    for (int j = 0 ; j < ref_data_list[p].size() ; j++) {
		ReferenceData ref_data = (ReferenceData)ref_data_list[p].elementAt(j);
		if (ref_data.getName().equals(data.getName())) {
		    ref_data_list[p].removeElementAt(j);
		    if (mode != 2)
			ref_data_list[p].insertElementAt(data, j);

		    replaced_flag = true;
		    break;
		}
	    }

	    if (replaced_flag == false  &&  mode != 2)
		ref_data_list[p].addElement(data);
	}

	modified_flag_list[p] = true;
    }

    public int loadableIndex ( ) {
	int data_size = 0;
	for (int i = 0 ; i < N ; i++) {
	    if (ref_file_list[i] != null  &&  ref_file_list[i].length() > 0)
		data_size += ref_data_list[i].size();
	}
	if (data_size < MAX_DATA_SIZE) {
	    for (int i = 0 ; i < N ; i++) {
		if (ref_file_list[i] == null  ||  ref_file_list[i].length() == 0)
		    return i;
	    }
	}

	int min = N + 1;
	int min_index = -1;
	int since_last_used_phase = -1;
	for (int i = 0 ; i < N ; i++) {
	    if (ref_file_list[i] != null  &&  ref_file_list[i].length() > 0) {
		int used_count = 0;
		for (int j = 0 ; j < N ; j++) {
		    if (used_flag_list[i][j])
			used_count++;
		}

		if (used_count < min  ||  used_count == 0) {
		    min = used_count;
		    min_index = i;
		    int j = 0;
		    for (j = 0 ; j < N ; j++) {
			int ph = used_flag_phase - j;
			if (ph < 0)
			    ph += N;
			if (used_flag_list[i][ph])
			    break;
		    }
		    since_last_used_phase = j;
		} else if (used_count == min) {
		    int j = 0;
		    for (j = 0 ; j < N ; j++) {
			int ph = used_flag_phase - j;
			if (ph < 0)
			    ph += N;
			if (used_flag_list[i][ph])
			    break;
		    }
		    if (j > since_last_used_phase) {
			since_last_used_phase = j;
			min_index = i;
		    }
		}
	    }
	}
	return min_index;
    }

    private void load ( int index, String filename ) {
	try {
	    save(index);
	} catch (Exception e) {
	    System.err.println("Cannot save file: " + ref_file_list[index]);
	}

	ref_data_list[index] = new Vector();
	ref_file_list[index] = filename;
	modified_flag_list[index] = false;

	try {
	    File file = new File(filename);
	    FileInputStream fis = new FileInputStream(file);
	    DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
	    String s = dis.readLine();
	    while (s != null) {
		StringTokenizer st = new StringTokenizer(s, "\t");
		String name = st.nextToken();
		String dbvs_name = st.nextToken();
		double ra = Float.valueOf(st.nextToken()).floatValue();
		double decl = Float.valueOf(st.nextToken()).floatValue();
		double pos_error = Float.valueOf(st.nextToken()).floatValue();
		ReferenceType type = ReferenceType.getType(st.nextToken().charAt(0));
		String record = dis.readLine();

		Coor cr = new Coor(ra, decl);
		ReferenceData data = new ReferenceData(name, dbvs_name, cr, record, pos_error, type);
		ref_data_list[index].addElement(data);

		s = dis.readLine();
	    }
	    dis.close();
	} catch (IOException e) {
	    System.err.println("Cannot load file: " + filename);
	}
    }

    public void save ( ) throws Exception {
	for (int i = 0 ; i < N ; i++)
	    save(i);
    }

    private void save ( int index ) throws Exception {
	if (ref_file_list[index] != null  &&  ref_file_list[index].length() > 0) {
	    if (modified_flag_list[index]) {
		FileOutputStream os = new FileOutputStream(ref_file_list[index]);
		PrintStream ps = new PrintStream(os);

		for (int j = 0 ; j < ref_data_list[index].size() ; j++) {
		    ReferenceData data = (ReferenceData)ref_data_list[index].elementAt(j);
		    ps.println(data.toString() + "\t" + Astro.toString(data.getPositionError(), 5, 3).trim() + "\t" + data.getType().toString());
		    ps.println(data.getRecord());
		}

		ps.close();

		modified_flag_list[index] = false;
	    }
	}
    }
}
