/*
 * PixyKernel.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.pixy;

import java.awt.Point;
import java.util.*;
import java.net.*;
import misao.util.*;
import misao.pixy.test.*;
import misao.diff.*;

public class PixyKernel {
    private PixyInputData input_data;
    private CatalogManager catalog_manager;

    private PrintManager print_manager;
    private boolean output_to_console;

    private boolean output_PS_data;

    TimeKeeper timer;
    private boolean timer_flag;

    private boolean test_detect;
    private boolean test_matching;

    // -test matching
    DetectList test_matching_matching_detect_list;
    DetectList test_matching_check_detect_list;
    CatalogList test_matching_matching_catalog_list;
    TriGraph test_matching_detect_graph;
    TriStarsIndex test_matching_tri_detect_index;
    TriGraph test_matching_catalog_graph;
    TriStarsIndex test_matching_tri_catalog_index;

    public PixyKernel ( PixyInputData data ) {
	input_data = data;
	catalog_manager = new CatalogManager(input_data);

	print_manager = new PrintManager();
	output_to_console = true;

	output_PS_data = true;

	timer_flag = false;

	test_detect = false;
	test_matching = false;
    }

    public void outputToConsole ( boolean f ) {
	output_to_console = f;
    }

    public void outputPSData ( boolean f ) {
	output_PS_data = f;
    }

    public void setTimer ( boolean f ) {
	timer_flag = f;
    }

    private void print_timer ( String s ) {
	if (timer != null  &&  timer_flag)
	    print_manager.println(s + "\t" + timer.check());
    }

    public void setExceptionThrowableAtNoCdrom ( boolean f ) {
	catalog_manager.setExceptionThrowableAtNoCdrom(f);
    }

    public void testDetect ( ) {
	test_detect = true;
    }

    public void testMatching ( ) {
	test_matching = true;
    }

    // main process of PIXY system
    public PixyOutputDataInterface operate ( ) throws Exception {
	PixyOutputData output_data = new PixyOutputData();

	ImageInformation image_info = output_data.image_info;
	image_info.image_url = input_data.image_url_string;
	image_info.exposure_start = input_data.exposure_start;
	image_info.exposure_seconds = input_data.exposure_seconds;
	image_info.mag_code = input_data.mag_code;

	print_manager.outputToConsole(output_to_console);
	print_manager.openFile(input_data.output_filename);

	if (print_manager.getPrintStreamOfFile() != null)
	    Version.printCopyright(print_manager.getPrintStreamOfFile(), "PIXY system");

	print_manager.println("**********************************************************************");
	print_manager.printObject(input_data);
	print_manager.println("**********************************************************************");

	Date date_start = new Date();

	timer = new TimeKeeper();

	// load the image
	ImagePixy image = new ImagePixy(new ImageFile(input_data.image_url).getGrayImage());
	output_data.image = image;
	image_info.image_size = image.getSize();
	print_manager.println(image.getSize().toString());
	print_manager.println(image.calculateMinMaxAverage().toString());
	print_timer("image loading");

	// coefficient to convert coordinate in pixel to deg
	double pixel_to_deg = input_data.fov / (double)image.getWidth();

	// reverse in case of a mirror image
	if (input_data.east_right)
	    image.mirrorHorizontal();

	// rescale the image in case of SBIG image (ST-4/6)
	if (input_data.sbig_ratio) {
	    int h = (int)((double)image.getHeight() * Astro.SBIG_RATIO + 0.5);
	    image = new ImagePixy(image.rescale(new Size(image.getWidth(), h)));
	    print_manager.println("SBIG ST-4/6 image. Rescaled: " + image.getSize().toString());
	}

	// detect stars
	DetectList detect_list = detect(image);

	// set mark on possible cosmic rays
	if (input_data.reject_cosmic_ray)
	    detect_list.markCosmicRay(0.4);

	// test display to check raw detected stars
	if (test_detect) {
	    // (x,y) is offset from the top-left corner
	    detect_list.shiftXY((double)image.getWidth() / 2.0, (double)image.getHeight() / 2.0);

	    PixyOutputDataDetect output_data_detect = new PixyOutputDataDetect();
	    output_data_detect.image = image;
	    output_data_detect.detect_list = detect_list;
	    return output_data_detect;
	}

	// catalog field of view for matching
	double fov = input_data.fov;
	if (image.getHeight() > image.getWidth())
	    fov *= (double)image.getHeight() / (double)image.getWidth();
	double catalog_fov = fov + input_data.position_error * 2.0;
	if (catalog_fov < fov * 2.0)
	    catalog_fov = fov * 2.0;
	print_manager.println("Search area size: " + catalog_fov + "x" + catalog_fov + " deg");

	// load star data for matching
	image_info.limiting_mag = Astro.getProperLimitingMagnitude(fov) - 2.0;
	print_manager.print("Loading catalog data ( " + image_info.limiting_mag + " mag, " + catalog_fov + " deg ) ...");
	CatalogList catalog_list = null;
	try {
	    catalog_list = catalog_manager.loadCatalogNull(input_data.center_coor, catalog_fov, image_info.limiting_mag);
	} catch ( Exception e ) {
	    print_manager.println(e.getMessage());
	    throw e;
	}
	catalog_list.sortMagAscendant();
	print_manager.println("done.");
	print_manager.println("" + catalog_list.size() + " stars found in catalog. (" + image_info.limiting_mag + " mag)");
	print_timer("loading catalog data for matching");

	// convert brightness of detected stars into magnitude
	// calculate rough magnitude translation formula
	int applied_count = detect_list.applyMagnitude(catalog_list, (int)((catalog_fov / fov) * (catalog_fov / fov) + 0.5), input_data.mag_code);
	output_data.pair_list = new PixyPairList(new PairList(detect_list, detect_list, applied_count));
	MagnitudeTranslationFormula mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple();
	detect_list.translateMagnitude(mag_formula, input_data.mag_code);
	image_info.mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateErrorMag(mag_formula));
	print_manager.println("Faintest mag: " + Astro.toString(detect_list.elementAt(applied_count - 1).getMag(), 5, 2).trim() + " (" + applied_count + " stars are applied.)");
	print_manager.println(mag_formula.toString());
	print_manager.println("  (+/-" + image_info.mag_error.toString() + ")");
	print_manager.println("Faintest mag: " + Astro.toString(detect_list.elementAt(applied_count - 1).getMag(), 5, 2).trim());
	print_timer("applying magnitude");

	// rescale catalog list
	catalog_list.rescaleUnit(pixel_to_deg);

	// select detected data used in matching
	SelectStarListIndex matching_detect_index = new SelectStarListIndex(detect_list, new Position(- (double)image.getWidth() / 2.0, - (double)image.getHeight() / 2.0), new Position((double)image.getWidth() / 2.0, (double)image.getHeight() / 2.0), 3, 100);

	// select catalog data used in matching
	SelectStarListIndex matching_catalog_index = new SelectStarListIndex(catalog_list, new Position(-0.5 * catalog_fov / pixel_to_deg, -0.5 * catalog_fov / pixel_to_deg), new Position(0.5 * catalog_fov / pixel_to_deg, 0.5 * catalog_fov / pixel_to_deg), (int)(4 * catalog_fov / fov + 0.5), (int)(125 * (catalog_fov / fov) * (catalog_fov / fov) + 0.5));

	// matching
	output_data.map_function = matching(matching_detect_index, matching_catalog_index, image.getSize());

	// test display to check data used for matching
	if (test_matching) {
	    PixyOutputDataMatching output_data_matching = new PixyOutputDataMatching();
	    output_data_matching.image = image;
	    output_data_matching.faintest_mag = image_info.limiting_mag;
	    output_data_matching.detect_list = detect_list;
	    output_data_matching.matching_detect_list = test_matching_matching_detect_list;
	    output_data_matching.check_detect_list = test_matching_check_detect_list;
	    output_data_matching.tri_detect_index = test_matching_tri_detect_index;
	    output_data_matching.tri_graph_detect = test_matching_detect_graph;
	    output_data_matching.catalog_list = catalog_list;
	    output_data_matching.matching_catalog_list = test_matching_matching_catalog_list;
	    output_data_matching.tri_catalog_index = test_matching_tri_catalog_index;
	    output_data_matching.tri_graph_catalog = test_matching_catalog_graph;
	    return output_data_matching;
	}

	matching_detect_index = null;
	matching_catalog_index = null;

	// approximate R.A. and Decl. of the true center, approximate true field of view
	Position center_on_chart = output_data.map_function.map(new Position(0,0));
	center_on_chart.rescale(pixel_to_deg);
	CoorMapping center_coor = new CoorMapping(input_data.center_coor, center_on_chart);
	fov *= output_data.map_function.getRate();
	double original_pixel_to_deg = pixel_to_deg;
	pixel_to_deg *= output_data.map_function.getRate();
	double original_catalog_fov = catalog_fov;
	catalog_fov = fov * 1.5;
	print_manager.println("Approximate true center: " + center_coor.toString());
	print_manager.println("Approximate true field of view: " + (input_data.fov * output_data.map_function.getRate()) + " deg");

	// load star data deeply
	double load_limit = Astro.getProperLimitingMagnitude(fov) + 2.0;
	print_manager.print("Loading catalog data ( " + load_limit + " mag, " + catalog_fov + " deg ) ...");
	try {
	    catalog_list = catalog_manager.loadCatalog(center_coor, catalog_fov, load_limit);
	} catch ( Exception e ) {
	    print_manager.println(e.getMessage());
	    throw e;
	}
	catalog_list.sortMagAscendant();
	print_manager.println("done.");
	print_manager.println("" + catalog_list.size() + " stars found in catalog. (" + load_limit + " mag)");
	print_timer("loading catalog data deeply");

	// re-map catalog data in the same coordinates as at matching
	catalog_list.mapElement(center_coor, input_data.center_coor, original_catalog_fov);
	catalog_list.rescaleUnit(original_pixel_to_deg);

	// preliminary pairing
	pairingByMovingAverage("Preliminary", output_data, detect_list, catalog_list, null, 5.0, 4.0, 50, 0.6);

	// determine limiting magnitude
	image_info.limiting_mag = determineLimitmagMagnitude(output_data, null);
	if (Math.abs(input_data.limiting_mag - Astro.ILLEGAL_MAG) > 1.0)
	    image_info.limiting_mag = input_data.limiting_mag;
	output_data.pair_list.removeDataFainterThan(image_info.limiting_mag);
	double value_limit = mag_formula.reverseCalculate(image_info.limiting_mag);
	print_manager.println("Preliminary limiting mag: " + Astro.toString(image_info.limiting_mag, 5, 2));

	// re-calculate magnitude of detected stars based on pair list
	if (output_data.pair_list.getPairCountBrighterThan(image_info.limiting_mag - 1.0) >= 10)
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple(image_info.limiting_mag - 1.0);
	else if (output_data.pair_list.getPairCountBrighterThan(image_info.limiting_mag) >= 10)
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple(image_info.limiting_mag);
	else
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple();
	detect_list.translateMagnitude(mag_formula, input_data.mag_code);
	image_info.mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateErrorMag(mag_formula));
	print_manager.println(mag_formula.toString());
	print_manager.println("  (+/-" + image_info.mag_error.toString() + ")");

	// re-calculate limiting magnitude
	image_info.limiting_mag = mag_formula.calculate(value_limit);
	if (Math.abs(input_data.limiting_mag - Astro.ILLEGAL_MAG) > 1.0)
	    image_info.limiting_mag = input_data.limiting_mag;
	print_manager.println("Limiting mag: " + Astro.toString(image_info.limiting_mag, 5, 2));

	// re-map catalog data in the same coordinates as at loading deeply
	catalog_list.mapElement(input_data.center_coor, center_coor, catalog_fov);
	catalog_list.rescaleUnit(pixel_to_deg);

	// blending of close catalog data
	StarListIndex catalog_index = new StarListIndex(catalog_list, new Position(- (double)catalog_fov / 2.0 / pixel_to_deg, - (double)catalog_fov / 2.0 / pixel_to_deg), new Position((double)catalog_fov / 2.0 / pixel_to_deg, (double)catalog_fov / 2.0 / pixel_to_deg), 20, new MapFunction());
	catalog_index.blend(new MapFunction(), 1.414);
	StarList list = catalog_index.getAllList();
	catalog_list = new CatalogList();
	for (int i = 0 ; i < list.size() ; i++)
	    catalog_list.addElement(list.elementAt(i));
	catalog_list.setUnit(pixel_to_deg);
	catalog_list.sortMagAscendant();
	catalog_index = null;
	print_manager.println("" + catalog_list.size() + " stars remained in catalog after blending.");

	// re-calculate map function and distortion field
	calculateMapping("Preliminary", output_data);
	((PixyPairList)output_data.pair_list).setAllDetectCoor(output_data.map_function, center_coor, pixel_to_deg);
	image_info.pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition();
	print_manager.println("position error: " + Astro.toString(image_info.pos_error.getRA(), 8, 5).trim() + "\" " + Astro.toString(image_info.pos_error.getDecl(), 8, 5).trim() + "\"");

	// quasi pairing based on preliminary map function, true magnitude and preliminary distortion field
	double d_mag = image_info.mag_error.getDeltaMag();
	StarList pairing_detect_list = new StarList(detect_list);
	pairing_detect_list.limitMagAt(image_info.limiting_mag + 2.0 * d_mag);
	StarList pairing_catalog_list = new StarList(catalog_list);
	pairing_catalog_list.limitMagAt(image_info.limiting_mag + 2.0 * d_mag + 2.0);
	double d_pixel = (image_info.pos_error.getRA() + image_info.pos_error.getDecl()) / 2.0 / 3600.0 / pixel_to_deg;
	if (d_pixel < 1.0)
	    d_pixel = 1.0;
	if (d_mag < 1.0)
	    d_mag = 1.0;
	pairingByMagLimit("Quasi", output_data, pairing_detect_list, pairing_catalog_list, null, image_info.limiting_mag, d_pixel * 3.0, d_mag * 3.0);
	pairing_detect_list = null;
	pairing_catalog_list = null;

	// re-calculate map function and distortion field
	calculateMapping("Quasi", output_data);
	((PixyPairList)output_data.pair_list).setAllDetectCoor(output_data.map_function, center_coor, pixel_to_deg);
	image_info.pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition();
	print_manager.println("position error: " + Astro.toString(image_info.pos_error.getRA(), 8, 5).trim() + "\" " + Astro.toString(image_info.pos_error.getDecl(), 8, 5).trim() + "\"");

	// precise pairing (repeatedly)
	boolean repeat_flag = true;
	int repeat_count = 1;
	Position p0 = image_info.distortion_field.getLargestValue();
	double v0 = p0.getX() * p0.getX() + p0.getY() * p0.getY();
	while (repeat_flag) {
	    pairingByMagLimit("Precise (" + repeat_count + ")", output_data, detect_list, catalog_list, null, image_info.limiting_mag, d_pixel, 99.9);
	    calculateMapping("Precise (" + repeat_count + ")", output_data);

	    Position p = image_info.distortion_field.getLargestValue();
	    double v = p.getX() * p.getX() + p.getY() * p.getY();
	    if (v0 / 1.2 < v  &&  v < v0 * 1.2)
		repeat_flag = false;
	    v0 = v;

	    repeat_count++;
	    if (repeat_count >= 10)
		repeat_flag = false;
	}

	// final pairing (2 pass)
	pairingByMagLimit("Final (1st)", output_data, detect_list, catalog_list, null, image_info.limiting_mag, d_pixel, 99.9);
	PairList pair_list = new PairList(output_data.pair_list);
	pairingByMagLimit("Final (2nd)", output_data, new StarList(output_data.pair_list.createFirstNonpairList()), new StarList(output_data.pair_list.createSecondNonpairList()), null, image_info.limiting_mag, d_pixel * 3.0, d_mag * 3.0);
	for (int i = 0 ; i < pair_list.getPairCount() ; i++) {
	    PairData pair = pair_list.elementPairAt(i);
	    output_data.pair_list.addPairElement(pair.getFirstStarData(), pair.getSecondStarData());
	}
	pair_list = null;

	// accurate astrometry
	calculateMapping("Final", output_data);
	((PixyPairList)output_data.pair_list).setAllDetectCoor(output_data.map_function, center_coor, pixel_to_deg, image_info.distortion_field);
	image_info.pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition();
	image_info.distortion_field.setPrintInDetail(true);
	print_manager.printObject(image_info.distortion_field);

	// calculate position error
	((PixyPairList)output_data.pair_list).setAllDetectCoor(output_data.map_function, center_coor, pixel_to_deg, image_info.distortion_field);
	image_info.pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition();
	print_manager.println("position error: " + Astro.toString(image_info.pos_error.getRA(), 8, 5).trim() + "\" " + Astro.toString(image_info.pos_error.getDecl(), 8, 5).trim() + "\"");

	// remove catalog data out of the image, or at edge
	int count = ((PixyPairList)output_data.pair_list).removeCatalogNonpairDataOutOfImage(output_data.map_function, image_info.distortion_field, image.getSize(), 5.0);
	print_manager.println("" + count + " NC stars out of the image or at edge are removed.");

	// remove cosmic ray (only non-pair ones)
	if (input_data.reject_cosmic_ray) {
	    count = ((PixyPairList)output_data.pair_list).removeCosmicRay();
	    print_manager.println("" + count + " cosmic rays are removed.");
	}

	// blend NC stars close by other pairs
	count = ((PixyPairList)output_data.pair_list).blendCatalogStars(output_data.map_function, image_info.distortion_field);
	print_manager.println("" + count + " NC stars are blended.");

	// adjust systematic shift of magnitude
	MagnitudeAdjustment ma = ((PixyPairList)output_data.pair_list).adjustMagnitude();
	print_manager.printObject(ma);

	// faintest magnitude
	output_data.pair_list.sortPairListInFirstMag();
	image_info.faintest_mag = output_data.pair_list.elementPairAt(output_data.pair_list.getPairCount() - 1).getFirstStarData().getMag();
	print_manager.println("Faintest mag = " + image_info.faintest_mag);

	// limiting magnitude
	double old_lm = image_info.limiting_mag + ma.getDeltaMag(image_info.limiting_mag);
	image_info.limiting_mag = ((PixyPairList)output_data.pair_list).calculateLimitingMag(image_info.faintest_mag);
	if (Math.abs(input_data.limiting_mag - Astro.ILLEGAL_MAG) > 1.0)
	    image_info.limiting_mag = input_data.limiting_mag;
	print_manager.println("Limiting mag = " + image_info.limiting_mag);
	if (image_info.limiting_mag > old_lm) {
	    if (Math.abs(input_data.limiting_mag - Astro.ILLEGAL_MAG) <= 1.0) {
		image_info.limiting_mag = old_lm;
		print_manager.println("Limiting mag (old style) = " + image_info.limiting_mag);
	    }
	}

	// remove ND, NC fainter than limiting mag
	output_data.pair_list.removeFirstNonpairDataFainterThan(image_info.limiting_mag);
	output_data.pair_list.removeSecondNonpairDataFainterThan(image_info.limiting_mag);

	// calculate constant magnitude error (independent error on brightness)
	image_info.mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateConstantMagError(50));
	print_manager.println("constant mag error = " + image_info.mag_error.getDeltaMag() + " mag");

	// calculate brightness dependent magnitude error
	image_info.mag_error = ((PixyPairList)output_data.pair_list).calculateBrightnessDependentMagError(image_info.mag_error.getConstantError());
	print_manager.printObject(image_info.mag_error);

	// remarkable paired data whose position or magnitude differs
	count = ((PixyPairList)output_data.pair_list).selectRemarkableList(image_info.pos_error, image_info.mag_error);
	output_data.pair_list.sortRemarkableListInMag();
	print_manager.println("" + count + " stars are remarkable.");

	// remove nonsense blending of catalog data
	count = ((PixyPairList)output_data.pair_list).removeNonsenseBlendingCatalogData(image_info.mag_error.getConstantError());
	print_manager.println("" + count + " nonsense blended catalog data are removed.");

	// remove faint RP data because they are often just mistakes
	count = ((PixyPairList)output_data.pair_list).removeMispairData(image_info.limiting_mag, pixel_to_deg);
	print_manager.println("" + count + " faint RP data are removed as mis-pairs.");

	// set errors of RM, RP objects in memo
	((PixyPairList)output_data.pair_list).setRemarkableErrorInMemo(image_info.pos_error, pixel_to_deg);

	// set (x,y) of NC objects
	((PixyPairList)output_data.pair_list).createDetectDataOfNC(output_data.map_function, image_info.distortion_field, image.getSize());

	// regard ND stars close by other pairs as mistakes of close binary separation in catalog
//	count = ((PixyPairList)output_data.pair_list).pairingBinaryStars(image_info.pos_error, 2.0);

	// (x,y) is offset from the top-left corner
	detect_list = new DetectList(new StarList(output_data.pair_list.createFirstList()));
	detect_list.shiftXY((double)image.getWidth() / 2.0, (double)image.getHeight() / 2.0);
	if (input_data.detail_star_data)
	    detect_list.setPrintDetail();
	else
	    detect_list.setPrintSimple();

	output_data.pair_list.sortFirstNonpairListInMag();
	output_data.pair_list.sortSecondNonpairListInMag();

	// names and aliases
	output_data.pair_list.setAllNames();
	((PixyPairList)output_data.pair_list).setAllAlias();

	print_timer("astrometry, etc.");

	// output image mapping
	print_manager.println("Image mapping");
	print_manager.print("  ");
	printCorner(output_data, pixel_to_deg, center_coor, -1, -1);
	print_manager.print("                          ");
	printCorner(output_data, pixel_to_deg, center_coor,  1, -1);
	print_manager.println("");
	print_manager.print("                            ");
	printCorner(output_data, pixel_to_deg, center_coor,  0,  0);
	print_manager.println("");
	print_manager.print("  ");
	printCorner(output_data, pixel_to_deg, center_coor, -1,  1);
	print_manager.print("                          ");
	printCorner(output_data, pixel_to_deg, center_coor,  1,  1);
	print_manager.println("");

	// other output data
	Position df = image_info.distortion_field.getValue(new Position(0,0));
	Position pos = output_data.map_function.map(df);
	pos.rescale(pixel_to_deg);
	image_info.center_coor = new CoorMapping(center_coor, pos);
	image_info.field_of_view = (double)image.getWidth() * pixel_to_deg * output_data.map_function.getRate();
	image_info.pa = 360.0 - output_data.map_function.getTheta();

	// output PXF file to console
	print_manager.println("**********************************************************************");
	PxfFile pxf = new PxfFile(null, output_data.image_info, output_data.pair_list);
	output_data.pair_list.outputAllData(false);
	print_manager.printObject(pxf);

	// save PXF file
	pxf = new PxfFile(input_data, output_data.image_info, output_data.pair_list);
	output_data.pair_list.outputAllData(true);
	output_data.pair_list.outputPSData(output_PS_data);
	PrintManager pxf_manager = new PrintManager();
	pxf_manager.outputToConsole(false);
	pxf_manager.openFile(input_data.pxf_filename);
	pxf_manager.printObject(pxf);
	pxf_manager.close();

	Date date_end = new Date();
	print_manager.println(date_start.toString());
	print_manager.println(date_end.toString());

	print_manager.println("Please send all outputs to Seiichi Yoshida by E-mail (comet@aerith.net).");

	print_manager.close();

	return output_data;
    }

    // differential PIXY system
    public DifferentialPixyOutputData difference ( DifferentialPixyInputData input_data ) throws Exception {
	PixyOutputData output_data = new PixyOutputData();

	ImageInformation image_info = output_data.image_info;

	print_manager.outputToConsole(output_to_console);
	print_manager.openFile(input_data.output_filename);

	if (print_manager.getPrintStreamOfFile() != null)
	    Version.printCopyright(print_manager.getPrintStreamOfFile(), "PIXY system");

	print_manager.println("**********************************************************************");
	print_manager.printObject(input_data);
	print_manager.println("**********************************************************************");

	Date date_start = new Date();

	timer = new TimeKeeper();

	// load the images
	ImagePixy image1 = new ImagePixy(new ImageFile(new URL(new URLString(input_data.image_url1).toString())).getGrayImage());
	print_manager.println(image1.getSize().toString());
	print_manager.println(image1.calculateMinMaxAverage().toString());
	print_timer("image 1 loading");
	ImagePixy image2 = new ImagePixy(new ImageFile(new URL(new URLString(input_data.image_url2).toString())).getGrayImage());
	print_manager.println(image2.getSize().toString());
	print_manager.println(image2.calculateMinMaxAverage().toString());
	print_timer("image 2 loading");

	// detect stars
	DetectList detect_list1 = detect(image1);
	DetectList detect_list2 = detect(image2);

	// convert brightness of detected stars into magnitude
	// calculate rough magnitude translation formula
	MagnitudeTranslationFormula mag_formula = new MagnitudeTranslationFormula(20.0, -1.0);
	detect_list2.translateMagnitude(mag_formula, "");
	int applied_count = detect_list1.applyMagnitude(detect_list2, 1, "");
	output_data.pair_list = new PixyPairList(new PairList(detect_list1, detect_list1, applied_count));
	mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple();
	detect_list1.translateMagnitude(mag_formula, "");
	MagError mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateErrorMag(mag_formula));
	print_manager.println(mag_formula.toString());
	print_manager.println("  (+/-" + mag_error.toString() + ")");
	print_manager.println("Faintest mag: " + Astro.toString(detect_list1.elementAt(applied_count - 1).getMag(), 5, 2).trim());
	print_timer("applying magnitude");

	// select detected data used in matching
	SelectStarListIndex matching_detect_index1 = new SelectStarListIndex(detect_list1, new Position(- (double)image1.getWidth() / 2.0, - (double)image1.getHeight() / 2.0), new Position((double)image1.getWidth() / 2.0, (double)image1.getHeight() / 2.0), 3, 100);
	SelectStarListIndex matching_detect_index2 = new SelectStarListIndex(detect_list2, new Position(- (double)image2.getWidth() / 2.0, - (double)image2.getHeight() / 2.0), new Position((double)image2.getWidth() / 2.0, (double)image2.getHeight() / 2.0), 3, 100);

	// matching
	output_data.map_function = matching(matching_detect_index1, matching_detect_index2, image2.getSize());

	matching_detect_index1 = null;
	matching_detect_index2 = null;

	output_data.image = image2;

	// preliminary pairing
	pairingByMovingAverage("Preliminary", output_data, detect_list1, detect_list2, image2.getSize(), 5.0, 4.0, 50, 0.6);

	// determine limiting magnitude
	image_info.limiting_mag = determineLimitmagMagnitude(output_data, image2.getSize());
	output_data.pair_list.removeDataFainterThan(image_info.limiting_mag);
	double value_limit = mag_formula.reverseCalculate(image_info.limiting_mag);
	print_manager.println("Preliminary limiting mag: " + Astro.toString(image_info.limiting_mag, 5, 2));

	// re-calculate magnitude of detected stars based on pair list
	if (output_data.pair_list.getPairCountBrighterThan(image_info.limiting_mag - 1.0) >= 10)
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple(image_info.limiting_mag - 1.0);
	else if (output_data.pair_list.getPairCountBrighterThan(image_info.limiting_mag) >= 10)
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple(image_info.limiting_mag);
	else
	    mag_formula = ((PixyPairList)output_data.pair_list).calculateMagnitudeTranslationFormulaInPrinciple();
	detect_list1.translateMagnitude(mag_formula, "");
	image_info.mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateErrorMag(mag_formula));
	print_manager.println(mag_formula.toString());
	print_manager.println("  (+/-" + image_info.mag_error.toString() + ")");

	// re-calculate limiting magnitude
	image_info.limiting_mag = mag_formula.calculate(value_limit);
	print_manager.println("Limiting mag: " + Astro.toString(image_info.limiting_mag, 5, 2));

	// re-calculate map function and distortion field
	calculateMapping("Preliminary", output_data);
	Position pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition(output_data.map_function);
	print_manager.println("position error: (" + Astro.toString(pos_error.getX(), 8, 5).trim() + ", " + Astro.toString(pos_error.getY(), 8, 5).trim() + ") pixels");

	// quasi pairing based on preliminary map function, true magnitude and preliminary distortion field
	double d_mag = image_info.mag_error.getDeltaMag();
	StarList pairing_detect_list1 = new StarList(detect_list1);
	pairing_detect_list1.limitMagAt(image_info.limiting_mag + 2.0 * d_mag);
	StarList pairing_detect_list2 = new StarList(detect_list2);
	pairing_detect_list2.limitMagAt(image_info.limiting_mag + 2.0 * d_mag);
	double d_pixel = Math.sqrt(pos_error.getX() * pos_error.getX() + pos_error.getY() * pos_error.getY());
	if (d_pixel < 1.0)
	    d_pixel = 1.0;
	if (d_mag < 1.0)
	    d_mag = 1.0;
	pairingByMagLimit("Quasi", output_data, pairing_detect_list1, pairing_detect_list2, image2.getSize(), image_info.limiting_mag, d_pixel * 3.0, d_mag * 3.0);
	pairing_detect_list1 = null;
	pairing_detect_list2 = null;

	// re-calculate map function and distortion field
	calculateMapping("Quasi", output_data);
	pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition(output_data.map_function);
	print_manager.println("position error: (" + Astro.toString(pos_error.getX(), 8, 5).trim() + ", " + Astro.toString(pos_error.getY(), 8, 5).trim() + ") pixels");

	// precise pairing (repeatedly)
	boolean repeat_flag = true;
	int repeat_count = 1;
	Position p0 = image_info.distortion_field.getLargestValue();
	double v0 = p0.getX() * p0.getX() + p0.getY() * p0.getY();
	while (repeat_flag) {
	    pairingByMagLimit("Precise (" + repeat_count + ")", output_data, detect_list1, detect_list2, image2.getSize(), image_info.limiting_mag, d_pixel, 99.9);
	    calculateMapping("Precise (" + repeat_count + ")", output_data);

	    Position p = image_info.distortion_field.getLargestValue();
	    double v = p.getX() * p.getX() + p.getY() * p.getY();
	    if (v0 / 1.2 < v  &&  v < v0 * 1.2)
		repeat_flag = false;
	    v0 = v;

	    repeat_count++;
	    if (repeat_count >= 10)
		repeat_flag = false;
	}

	// final pairing (2 pass)
	pairingByMagLimit("Final (1st)", output_data, detect_list1, detect_list2, image2.getSize(), image_info.limiting_mag, d_pixel, 99.9);
	PairList pair_list = new PairList(output_data.pair_list);
	pairingByMagLimit("Final (2nd)", output_data, new StarList(output_data.pair_list.createFirstNonpairList()), new StarList(output_data.pair_list.createSecondNonpairList()), image2.getSize(), image_info.limiting_mag, d_pixel * 3.0, d_mag * 3.0);
	for (int i = 0 ; i < pair_list.getPairCount() ; i++) {
	    PairData pair = pair_list.elementPairAt(i);
	    output_data.pair_list.addPairElement(pair.getFirstStarData(), pair.getSecondStarData());
	}
	pair_list = null;

	// accurate astrometry
	calculateMapping("Final", output_data);
	pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition(output_data.map_function);
	image_info.distortion_field.setPrintInDetail(true);
	print_manager.printObject(image_info.distortion_field);

	// calculate position error
	pos_error = ((PixyPairList)output_data.pair_list).calculateErrorPosition(output_data.map_function);
	print_manager.println("position error: (" + Astro.toString(pos_error.getX(), 8, 5).trim() + ", " + Astro.toString(pos_error.getY(), 8, 5).trim() + ") pixels");

	// remove data out of the other image, or at edge
	int count = ((PixyPairList)output_data.pair_list).removeDetectNonpairDataOutOfImage(output_data.map_function, image_info.distortion_field, image2.getSize(), 5.0);
	print_manager.println("" + count + " stars out of the image or at edge are removed in image1.");
	count = ((PixyPairList)output_data.pair_list).removeCatalogNonpairDataOutOfImage(output_data.map_function, image_info.distortion_field, image1.getSize(), 5.0);
	print_manager.println("" + count + " stars out of the image or at edge are removed in image2.");

	// adjust systematic shift of magnitude
	MagnitudeAdjustment ma = ((PixyPairList)output_data.pair_list).adjustMagnitude();
	print_manager.printObject(ma);

	// faintest magnitude
	output_data.pair_list.sortPairListInFirstMag();
	image_info.faintest_mag = output_data.pair_list.elementPairAt(output_data.pair_list.getPairCount() - 1).getFirstStarData().getMag();
	print_manager.println("Faintest mag = " + image_info.faintest_mag);

	// limiting magnitude
	double old_lm = image_info.limiting_mag + ma.getDeltaMag(image_info.limiting_mag);
	image_info.limiting_mag = ((PixyPairList)output_data.pair_list).calculateLimitingMag(image_info.faintest_mag);
	print_manager.println("Limiting mag = " + image_info.limiting_mag);
	if (image_info.limiting_mag > old_lm) {
	    image_info.limiting_mag = old_lm;
	    print_manager.println("Limiting mag (old style) = " + image_info.limiting_mag);
	}

	// remove ND, NC fainter than limiting mag
	output_data.pair_list.removeFirstNonpairDataFainterThan(image_info.limiting_mag);
	output_data.pair_list.removeSecondNonpairDataFainterThan(image_info.limiting_mag);

	// calculate constant magnitude error (independent error on brightness)
	image_info.mag_error = new MagError(((PixyPairList)output_data.pair_list).calculateConstantMagError(50));
	print_manager.println("constant mag error = " + image_info.mag_error.getDeltaMag() + " mag");

	// calculate brightness dependent magnitude error
	image_info.mag_error = ((PixyPairList)output_data.pair_list).calculateBrightnessDependentMagError(image_info.mag_error.getConstantError());
	print_manager.printObject(image_info.mag_error);

	// remarkable paired data whose position or magnitude differs
	count = ((PixyPairList)output_data.pair_list).selectRemarkableList(pos_error, image_info.mag_error, output_data.map_function);
	output_data.pair_list.sortRemarkableListInMag();
	print_manager.println("" + count + " stars are remarkable.");

	// calculate error at limiting mag
	double dmag_at_lm = image_info.mag_error.getRemarkableThreshold(image_info.limiting_mag);
	double dmag_at_lm2 = image_info.mag_error.getRemarkableThreshold(image_info.limiting_mag - dmag_at_lm);
	dmag_at_lm = (dmag_at_lm + dmag_at_lm2) / 2.0;
	print_manager.println("Mag error at limiting mag = " + dmag_at_lm);

	// threshold mag to output RP and NP
	double limit_to_output = image_info.limiting_mag - dmag_at_lm;
	print_manager.println("Threshold mag to output RP and NP = " + limit_to_output);

	// remove RM and NP fainter than limiting mag
	output_data.pair_list.removeRemarkableDataFainterThan(limit_to_output);
	output_data.pair_list.removeFirstNonpairDataFainterThan(limit_to_output);
	output_data.pair_list.removeSecondNonpairDataFainterThan(limit_to_output);

	// set errors of RM, RP objects in memo
	((PixyPairList)output_data.pair_list).setRemarkableErrorInMemo(pos_error, output_data.map_function);

	// set (x,y) in the other image of ND, NC objects
	((PixyPairList)output_data.pair_list).createDetectDataOfND(output_data.map_function, image_info.distortion_field, image2.getSize());
	((PixyPairList)output_data.pair_list).createDetectDataOfNC(output_data.map_function, image_info.distortion_field, image1.getSize());

	// (x,y) is offset from the top-left corner
	detect_list1 = new DetectList(new StarList(output_data.pair_list.createFirstList()));
	detect_list1.shiftXY((double)image1.getWidth() / 2.0, (double)image1.getHeight() / 2.0);
	detect_list1.setPrintOnlyXYMag();
	detect_list2 = new DetectList(new StarList(output_data.pair_list.createSecondList()));
	detect_list2.shiftXY((double)image2.getWidth() / 2.0, (double)image2.getHeight() / 2.0);
	detect_list2.setPrintOnlyXYMag();

	// measure checking magnitude in 5x5 pixels
	measureCheckingMagnitude(output_data.pair_list.createFirstList(true), image1, mag_formula, 3);
	measureCheckingMagnitude(output_data.pair_list.createSecondList(true), image2, new MagnitudeTranslationFormula(20.0, -1.0), 3);

	// remove illegal differential data based on the checking magnitude
	((PixyPairList)output_data.pair_list).removeIllegalDifferentialPairs(image_info.mag_error, image1.getSize(), image2.getSize());

	// sort
	output_data.pair_list.sortFirstNonpairListInMag();
	output_data.pair_list.sortSecondNonpairListInMag();

	// overlapping area ratio
	double x, y;
	int percent = 0;
	count = 0;
	for (y = - (double)image1.getHeight() / 2.05 ; y <= (double)image1.getHeight() / 2.05 ; y += (double)image1.getHeight() / 2.05 / 10.01) {
	    for (x = - (double)image1.getWidth() / 2.05 ; x <= (double)image1.getWidth() / 2.05 ; x += (double)image1.getWidth() / 2.05 / 10.01) {
		count++;
		Position pos = output_data.map_function.map(new Position(x,y));
		if (- image2.getWidth() / 2 <= pos.getX()  &&  pos.getX() < image2.getWidth() / 2  &&
		    - image2.getHeight() / 2 <= pos.getY()  &&  pos.getY() < image2.getHeight() / 2)
		    percent++;
	    }
	}
	double overlap_ratio = (double)percent / (double)count;

	print_timer("astrometry, etc.");

	// output data
	DifferentialPixyOutputData diff_output_data = new DifferentialPixyOutputData();
	diff_output_data.image1 = image1;
	diff_output_data.image2 = image2;
	diff_output_data.image_info.image_url1 = input_data.image_url1;
	diff_output_data.image_info.image_url2 = input_data.image_url2;
	diff_output_data.image_info.image_size1 = image1.getSize();
	diff_output_data.image_info.image_size2 = image2.getSize();
	diff_output_data.image_info.pos_error = pos_error;
	diff_output_data.image_info.mag_error = image_info.mag_error;
	diff_output_data.image_info.limiting_mag = image_info.limiting_mag;
	diff_output_data.image_info.faintest_mag = image_info.faintest_mag;
	diff_output_data.image_info.overlap_ratio = overlap_ratio;
	diff_output_data.pair_list = new PairList(output_data.pair_list);
	diff_output_data.map_function = output_data.map_function;

	// names
	diff_output_data.pair_list.setAllNames();

	// output PXF file to console
	print_manager.println("**********************************************************************");
	DifferentialPxfFile pxf = new DifferentialPxfFile(diff_output_data.image_info, diff_output_data.pair_list);
	diff_output_data.pair_list.outputAllData(false);
	print_manager.printObject(pxf);

	// save PXF file
	diff_output_data.pair_list.outputAllData(true);
	diff_output_data.pair_list.outputPSData(output_PS_data);
	PrintManager pxf_manager = new PrintManager();
	pxf_manager.outputToConsole(false);
	pxf_manager.openFile(input_data.pxf_filename);
	pxf_manager.printObject(pxf);
	pxf_manager.close();

	Date date_end = new Date();
	print_manager.println(date_start.toString());
	print_manager.println(date_end.toString());

	print_manager.println("Please send all outputs to Seiichi Yoshida by E-mail (comet@aerith.net).");

	print_manager.close();

	return diff_output_data;
    }

    private void printCorner ( PixyOutputData output_data, double pixel_to_deg, CoorMapping center_coor, int dw, int dh ) {
	Position pos = new Position((double)(dw * output_data.image.getWidth()) / 2.0, (double)(dh * output_data.image.getHeight()) / 2.0);
	Position df_pos = output_data.image_info.distortion_field.getValue(pos);
	pos = output_data.map_function.map(new Position(pos.getX() + df_pos.getX(), pos.getY() + df_pos.getY()));
	pos.rescale(pixel_to_deg);
	Coor coor = new CoorMapping(center_coor, pos);
	print_manager.print(coor.toStringRA() + "  " + coor.toStringDecl());
    }

    // determine limiting magnitude
    private double determineLimitmagMagnitude ( PixyOutputData output_data, Size catalog_size ) {
	int limitmag_data_count = 0;
	StarList limitmag_pair_detect_list = new StarList(output_data.pair_list.createFirstPairList());
	for (int i = 0 ; i < limitmag_pair_detect_list.size() ; i++) {
	    DetectData sd_detect = (DetectData)limitmag_pair_detect_list.elementAt(i);
	    if (! sd_detect.isEdge())
		limitmag_data_count++;
	}
	StarList limitmag_nonpair_detect_list = new StarList(output_data.pair_list.createFirstNonpairList());
	for (int i = 0 ; i < limitmag_nonpair_detect_list.size() ; i++) {
	    DetectData sd_detect = (DetectData)limitmag_nonpair_detect_list.elementAt(i);
	    Position mapped_pos = output_data.map_function.map(sd_detect.getPosition());
	    if (! sd_detect.isEdge()) {
		// not in case of out of the other image
		if (catalog_size == null  ||  
		    (- catalog_size.getWidth() / 2 <= mapped_pos.getX()  &&  mapped_pos.getX() <= catalog_size.getWidth() / 2  &&
		     - catalog_size.getHeight() / 2 <= mapped_pos.getY()  &&  mapped_pos.getY() <= catalog_size.getHeight() / 2)) {
		    limitmag_data_count++;
		}
	    }
	}
	int limitmag_index = limitmag_data_count;
	Array limitmag_array = new Array(limitmag_data_count);
	Array limitmag_array_mag = new Array(limitmag_data_count);
	DetectData sd_detect_pair = (DetectData)limitmag_pair_detect_list.getFirstElement();
	DetectData sd_detect_nonpair = (DetectData)limitmag_nonpair_detect_list.getFirstElement();
	int average_window_size = 19;
	double threshold_percent = 0.8;
	int count = 0;
	while (sd_detect_pair != null  ||  sd_detect_nonpair != null) {
	    while (sd_detect_pair != null  &&  sd_detect_pair.isEdge())
		sd_detect_pair = (DetectData)limitmag_pair_detect_list.getNextElement();
	    boolean nonpair_flag = true;
	    while (nonpair_flag) {
		if (sd_detect_nonpair == null) {
		    nonpair_flag = false;
		} else if (sd_detect_nonpair.isEdge()) {
		    sd_detect_nonpair = (DetectData)limitmag_nonpair_detect_list.getNextElement();
		} else {
		    Position mapped_pos = output_data.map_function.map(sd_detect_nonpair.getPosition());
		    // not in case of out of the other image
		    if (catalog_size == null  ||  
			(- catalog_size.getWidth() / 2 <= mapped_pos.getX()  &&  mapped_pos.getX() <= catalog_size.getWidth() / 2  &&
			 - catalog_size.getHeight() / 2 <= mapped_pos.getY()  &&  mapped_pos.getY() <= catalog_size.getHeight() / 2)) {
			nonpair_flag = false;
		    } else {
			sd_detect_nonpair = (DetectData)limitmag_nonpair_detect_list.getNextElement();
		    }
		}
	    }
	    if (sd_detect_pair == null  &&  sd_detect_nonpair == null)
		break;
	    if (sd_detect_nonpair == null) {
		limitmag_array.set(count, 1.0);
		limitmag_array_mag.set(count, sd_detect_pair.getMag());
		sd_detect_pair = (DetectData)limitmag_pair_detect_list.getNextElement();
	    } else if (sd_detect_pair == null) {
		limitmag_array.set(count, 0.0);
		limitmag_array_mag.set(count, sd_detect_nonpair.getMag());
		sd_detect_nonpair = (DetectData)limitmag_nonpair_detect_list.getNextElement();
	    } else if (sd_detect_pair.getMag() < sd_detect_nonpair.getMag()) {
		limitmag_array.set(count, 1.0);
		limitmag_array_mag.set(count, sd_detect_pair.getMag());
		sd_detect_pair = (DetectData)limitmag_pair_detect_list.getNextElement();
	    } else {
		limitmag_array.set(count, 0.0);
		limitmag_array_mag.set(count, sd_detect_nonpair.getMag());
		sd_detect_nonpair = (DetectData)limitmag_nonpair_detect_list.getNextElement();
	    }
	    count++;
	}
	if (limitmag_data_count >= average_window_size) {
	    limitmag_array.calculateMovingAverage(average_window_size);
	    limitmag_array.calculateMovingAverage(average_window_size);

	    int peak_index = limitmag_array.getPeakIndex();
	    limitmag_array.removeElement(0, peak_index);

	    limitmag_index = limitmag_array.meanIndexOf(threshold_percent);
	    if (limitmag_index < 0) {
		if (limitmag_array.getValue(0) >= threshold_percent)
		    limitmag_index = limitmag_data_count - 1;
		else
		    limitmag_index = average_window_size;
	    } else {
		limitmag_index = limitmag_array.meanIndexOf(threshold_percent) + peak_index;
		if (limitmag_index < average_window_size)
		    limitmag_index = average_window_size;
	    }
	}
	return limitmag_array_mag.getValue(limitmag_index);
    }

    // pairing based on the previous map function and distortion field
    private void pairingByMovingAverage ( String title, PixyOutputData output_data, StarList detect_list, StarList catalog_list, Size catalog_size, double d_pixel, double d_mag, int moving_average_size, double moving_average_threshold ) {
	pairing(title, output_data, detect_list, catalog_list, catalog_size, true, 99.9, d_pixel, d_mag, moving_average_size, moving_average_threshold);
    }

    private void pairingByMagLimit ( String title, PixyOutputData output_data, StarList detect_list, StarList catalog_list, Size catalog_size, double mag_limit, double d_pixel, double d_mag ) {
	pairing(title, output_data, detect_list, catalog_list, catalog_size, false, mag_limit, d_pixel, d_mag, 1, -1.0);
    }

    private void pairing ( String title, PixyOutputData output_data, StarList detect_list, StarList catalog_list, Size catalog_size, boolean pair_with_brightest, double mag_limit, double d_pixel, double d_mag, int moving_average_size, double moving_average_threshold ) {
	print_manager.print(title + " pairing...");
	output_data.pair_list = new PixyPairList();
	MovingAverage mov_ave = new MovingAverage(moving_average_size);

	StarList pairing_detect_list = new StarList(detect_list);
	pairing_detect_list.limitMagAt(mag_limit + 2.0 * output_data.image_info.mag_error.getDeltaMag());
	StarListIndex pairing_catalog_index = new StarListIndex(catalog_list, new Position(- (double)output_data.image.getWidth() / 2.0, - (double)output_data.image.getHeight() / 2.0), new Position((double)output_data.image.getWidth() / 2.0, (double)output_data.image.getHeight() / 2.0), 20, output_data.map_function.inverse());

	while (pairing_detect_list.size() > 0) {
	    StarData sd_detect = pairing_detect_list.elementAt(0);
	    Position df_pos = new Position(output_data.image_info.distortion_field.getValue(sd_detect.getPosition()));

	    // find a counterpart within the nearby area
	    // the found data will be removed from pairing_catalog_index
	    StarData sd_catalog = null;
	    if (pair_with_brightest)
		sd_catalog = pairing_catalog_index.extractPairBrightest(sd_detect, output_data.map_function.inverse(), df_pos, d_pixel, d_mag);
	    else
		sd_catalog = pairing_catalog_index.extractPair(sd_detect, output_data.map_function.inverse(), df_pos, d_pixel, d_mag);

	    if (sd_catalog != null) {
		output_data.pair_list.addPairElement(sd_detect, sd_catalog);
		mov_ave.add(1);
	    } else {
		output_data.pair_list.addFirstNonpairElement(sd_detect);

		// not in case of out of the other image
		Position mapped_pos = output_data.map_function.map(sd_detect.getPosition());
		if (catalog_size == null  ||  
		    (- catalog_size.getWidth() / 2 <= mapped_pos.getX()  &&  mapped_pos.getX() <= catalog_size.getWidth() / 2  &&
		     - catalog_size.getHeight() / 2 <= mapped_pos.getY()  &&  mapped_pos.getY() <= catalog_size.getHeight() / 2)) {
		    mov_ave.add(0);
		}
	    }
	    pairing_detect_list.removeElementAt(0);

	    if (mov_ave.getCount() >= moving_average_size  &&  mov_ave.getAverage() < moving_average_threshold)
		break;
	}
	output_data.pair_list.appendFirstNonpairList(pairing_detect_list);
	output_data.pair_list.appendSecondNonpairList(pairing_catalog_index.getAllList());
	pairing_detect_list = null;
	pairing_catalog_index = null;
	print_manager.println("done.");
	print_manager.println("search area size: " + d_pixel + " pixel, " + d_mag + " mag");
	print_manager.println("" + output_data.pair_list.getPairCount() + " stars are paired.");
	print_timer(title.toLowerCase() + " pairing");
    }

    // calculate map function and distortion field
    public void calculateMapping ( String title, PixyOutputData output_data ) throws Exception {
	// re-calculate map function based on pair list
	output_data.map_function = ((PixyPairList)output_data.pair_list).calculateMapFunction(output_data.image_info.distortion_field);
	print_manager.println(title + " map function:");
	print_manager.println(output_data.map_function.toString());

	// distortion field
	output_data.image_info.distortion_field = new DistortionField(output_data.pair_list, output_data.map_function);
	output_data.image_info.distortion_field.setPrintSize(output_data.image.getSize());
	print_manager.println(title + " distortion field:");
	print_manager.printObject(output_data.image_info.distortion_field);
	print_timer(title.toLowerCase() + " mapping");
    }

    // detect stars
    private DetectList detect ( ImagePixy image ) throws Exception {
	// equalize level of every row and column
	// roll vertical or horizontal white/dark line, especially being at edges
	image.equalizeRowColumn();

	// sky field
	print_manager.print("Calculating sky field...");
	ImagePixy sky_image = new ImagePixy(image.rescale(new Size(image.getWidth() / 8, image.getHeight() / 8)));
	sky_image = new ImagePixy(sky_image.median(5));
	double sky_value = 0.0;
/*
	if (input_data.image_flat) {
	    // flat sky field
	    MinMaxAverage mma_sky = sky_image.calculateMinMaxAverage();
	    sky_value = mma_sky.getAverage();
	    sky_image.fill(sky_value);
	    sky_image = new ImagePixy(sky_image.rescale(image.getSize()));
	} else {
*/
	    // sky field as a result of median filter
	    sky_image = new ImagePixy(sky_image.rescale(image.getSize()));
	    sky_image = new ImagePixy(sky_image.smooth_average(16));
	    MinMaxAverage mma_sky = sky_image.calculateMinMaxAverage();
	    sky_value = mma_sky.getAverage();
/*
	}
*/
	print_manager.println("done.");
	print_manager.println("sky field = " + sky_value);
	print_timer("sky field");

	// noise
	double noise = image.calculateNoise(sky_image) * 2.0;
	print_manager.println("noise = " + noise);

	// image of only stars
	ImagePixy stellar_image = image.stellar(sky_image, noise);
	print_timer("noise + stellar image");

	// in case of negative image
	int stellar_count = stellar_image.countBrighterThan(0.0);
	if (stellar_count < 100) {
	    // image may be negative
	    image.multiplyValue(-1.0);
	    stellar_image.multiplyValue(-1.0);
	    sky_value = - sky_value;

	    noise = image.calculateNoise(sky_image) * 2.0;
	    print_manager.println("noise = " + noise);

	    stellar_image = image.stellar(sky_image, noise);
	    int stellar_count2 = stellar_image.countBrighterThan(0.0);
	    if (stellar_count2 > stellar_count) {
		print_manager.println("Image is negative. Reversed. (stellar count: " + stellar_count + " -> " + stellar_count2 + ")");
	    } else {
		// image is not negative
		print_manager.println("Image is not negative. (stellar count: " + stellar_count + " -> " + stellar_count2 + ")");
		image.multiplyValue(-1.0);
		stellar_image.multiplyValue(-1.0);
		sky_value = - sky_value;
		stellar_image = image.stellar(sky_image, noise);
	    }
	    print_timer("reverse negative image");
	}

	// detect stars as adjoining pixels
	// stellar_image will be cleared
	AdjoiningPixelSet ap_set = stellar_image.detect();
	stellar_image = null;
	print_manager.println("" + ap_set.size() + " adjoining pixels are detected.");
	print_timer("detect adjoining pixels");

	// deblend adjoining pixels into stars
	// all elements of ap_set will be destroyed
	AdjoiningPixel ap = ap_set.getFirstElement();
	DetectList detect_list = new DetectList();
	while (ap != null) {
	    // ap will be destroyed
	    AdjoiningPixelSet ap_set1 = ap.deblend(noise);

	    AdjoiningPixel ap1 = ap_set1.getFirstElement();
	    while (ap1 != null) {
		detect_list.addElement(new DetectData(ap1, 9));
		ap1 = ap_set1.getNextElement();
	    }

	    ap = ap_set.getNextElement();
	}
	ap_set = null;
	print_manager.println("" + detect_list.size() + " stars are detected.");
	print_timer("deblend adjoining pixels");

	// PSF radius
	double radius = 0.0;
	double radius2 = 0.0;
	for (int i = 0 ; i < detect_list.size() ; i++) {
	    double r = ((DetectData)detect_list.elementAt(i)).getPSF().getRadius();
	    radius += r;
	    radius2 += r * r;
	}
	radius /= (double)detect_list.size();
	radius2 = Math.sqrt(radius2 / (double)detect_list.size() - radius * radius);
	print_manager.println("mean PSF radius = " + Astro.toString(radius, 6, 3).trim() + " (sigma = " + Astro.toString(radius2, 6, 3) + " )");

	// determine owner stars for each pixel
/*
	print_manager.print("Searching owner stars of each pixel...");
	PsfDistribution psf_distribute = new PsfDistribution(image.getSize());
	psf_distribute.owner(detect_list, 20.0);
	print_manager.println("done.");
	print_timer("search owner stars of each pixel");
*/

	// distribute pixel values
/*
	print_manager.print("Distributing pixel value...");
	detect_list.distribute(image, sky_image, psf_distribute);
	psf_distribute = null;
	sky_image = null;
	print_manager.println("done.");
	print_timer("distribute pixel value");
*/

	// (x,y) is offset from the center
	detect_list.shiftXY(- (double)image.getWidth() / 2.0, - (double)image.getHeight() / 2.0);
	detect_list.sortValueDescendant();

	return detect_list;
    }

    // matching
    private MapFunction matching ( SelectStarListIndex matching_detect_index, SelectStarListIndex matching_catalog_index, Size image_size ) throws MatchingFailedException {
	// select detected data used in matching
	DetectList matching_detect_list = new DetectList(matching_detect_index.select(3, 0));
	DetectList check_detect_list = new DetectList(matching_detect_index.select(9, 0));
	StarData sd = check_detect_list.getFirstElement();
	if (sd instanceof DetectData)
	    check_detect_list.removeElementAtEdge();
	print_manager.println("" + matching_detect_list.size() + " detected stars used in matching.");

	// index of trianles in detect list
	TriGraph detect_graph = new TriGraph(matching_detect_list);
	Vector tri_detect_list = detect_graph.createTriList(4);
	TriStarsIndex tri_detect_index = new TriStarsIndex();
	for (int i = 0 ; i < tri_detect_list.size() ; i++) {
	    TriStars tri_detect = (TriStars)tri_detect_list.elementAt(i);
	    tri_detect_index.addTriStars(tri_detect);
	}
	tri_detect_list = tri_detect_index.getAllList();
	print_manager.println("" + tri_detect_list.size() + " triangles in detected list.");

	// select catalog data used in matching
	CatalogList catalog_list = new CatalogList(matching_catalog_index.select(3, matching_catalog_index.getAllList().size()));
	CatalogList matching_catalog_list = new CatalogList(matching_catalog_index.select(3, 0));
	print_manager.println("" + matching_catalog_list.size() + " catalog stars used in matching.");

	// index of trianles in catalog list
//	int steps = (int)Math.sqrt(matching_catalog_list.size() * (fov / catalog_fov) * (fov / catalog_fov) * 2.0);
	TriGraph catalog_graph = new TriGraph(matching_catalog_list);
	Vector tri_catalog_list = catalog_graph.createTriList(8);
	TriStarsIndex tri_catalog_index = new TriStarsIndex();
	for (int i = 0 ; i < tri_catalog_list.size() ; i++) {
	    TriStars tri_catalog = (TriStars)tri_catalog_list.elementAt(i);
	    tri_catalog_index.addTriStars(tri_catalog);
	}
	print_manager.println("" + tri_catalog_list.size() + " triangles in catalog list.");

	// test display to check data used for matching
	if (test_matching) {
	    test_matching_matching_detect_list = matching_detect_list;
	    test_matching_check_detect_list = check_detect_list;
	    test_matching_matching_catalog_list = matching_catalog_list;
	    test_matching_detect_graph = detect_graph;
	    test_matching_tri_detect_index = tri_detect_index;
	    test_matching_catalog_graph = catalog_graph;
	    test_matching_tri_catalog_index = tri_catalog_index;
	    return null;
	}

	MapFunction map_function = new MapFunction();

	// matching
	// calculate candidates of map function repeatedly with similar triangles
	// map function: image -> chart
	print_manager.println("Matching start");
	double best_score = 0.0;
	for (int j = 0 ; j < tri_detect_list.size() ; j++) {
	    TriStars tri_detect = (TriStars)tri_detect_list.elementAt(j);
	    Point tri_pt = tri_catalog_index.findIndex(tri_detect);

	    tri_catalog_list = tri_catalog_index.findList(tri_detect);
	    print_manager.println("  " + j + " / " + tri_detect_list.size() + " : " + tri_catalog_list.size() + "  (" + tri_pt.x + ", " + tri_pt.y + ")");

	    for (int k = 0 ; k < tri_catalog_list.size() ; k++) {
		TriStars tri_catalog = (TriStars)tri_catalog_list.elementAt(k);

		MapFunctionCandidate mf = MapFunctionCandidate.calculate(tri_detect, tri_catalog);
		if (mf != null) {
		    // check if the map function is right and calculate score
		    if (0.5 < mf.getRate()  &&  mf.getRate() < 2.0) {
			CatalogList cl = new CatalogList(catalog_list, (int)((double)catalog_list.size() / mf.getRate()));
			double score = mf.checkValidity(check_detect_list, cl, image_size, check_detect_list.size());

			if (score > best_score) {
			    best_score = score;
			    map_function = mf;
			    print_manager.println(mf.toString());
			    print_manager.println("score: " + Astro.toString(score, 4, 1));

			    if (best_score > 0.5)
				break;
			}
		    }
		}
	    }

	    if (best_score > 0.5)
		break;
	}

	print_manager.println("Approximate map function:");
	print_manager.println(map_function.toString());
	print_timer("matching");
	if (best_score < 0.2) {
	    print_manager.println("Matching failed.");
	    throw new MatchingFailedException();
	}

	return map_function;
    }

    // measure checking magnitude in NxN pixels
    private void measureCheckingMagnitude ( Vector star_list, ImagePixy image, MagnitudeTranslationFormula mag_formula, int radius ) {
	int radius_sky = radius + 2;	// two lines for sky estimation

	for (int i = 0 ; i < star_list.size() ; i++) {
	    DetectData sd = (DetectData)star_list.elementAt(i);

	    // (x,y) is offset from the top-left corner
	    int x = (int)sd.getX();
	    int y = (int)sd.getY();

	    if (x < radius_sky  ||  sd.getX() >= image.getWidth() - radius_sky  ||
		y < radius_sky  ||  sd.getY() >= image.getHeight() - radius_sky) {
	    } else {
		double amount = 0;
		Point pt = new Point(0, 0);
		for (pt.y = - radius ; pt.y <= radius ; pt.y++) {
		    for (pt.x = - radius ; pt.x <= radius ; pt.x++) {
			amount += image.getValue(new Point(pt.x + x, pt.y + y));
		    }
		}

		int sky_size = (radius_sky*2+1) * (radius_sky*2+1) - (radius*2+1) * (radius*2+1);
		Array array = new Array(sky_size);
		int index = 0;
		for (pt.y = - radius_sky ; pt.y <= radius_sky ; pt.y++) {
		    for (pt.x = - radius_sky ; pt.x <= radius_sky ; pt.x++) {
			if (Math.abs(pt.y) > radius  ||  Math.abs(pt.x) > radius) {
			    array.set(index, image.getValue(new Point(pt.x + x, pt.y + y)));
			    index++;
			}
		    }
		}
		array.sortAscendant();

		double sky = array.getValue(sky_size / 2);
		amount -= (radius*2+1) * (radius*2+1) * sky;

		if (amount > 0)
		    sd.setCheckMag(mag_formula.calculate(amount));
	    }
	}
    }
}
