/*
 * CatalogCanvas.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.util;

import java.awt.*;
import java.util.*;

public class CatalogCanvas extends Canvas {
    // if parent frame is null, dialog will not open
    private Frame parent;

    private StarList catalog_list;
    private MapFunction map_function;
    private Image bg_image;
    private Size size;
    private Position base_pos;
    private Point mark_pt;
    private Color mark_color;
    private double mag_limit;
    private boolean mark_center;
    private boolean east_right;
    private boolean bg_black;
    private boolean frame_line;

    private boolean dialog_enable;	// true if you want to open star data dialog

    private boolean draw_foreground;
    private boolean draw_background;

    public CatalogCanvas ( Frame frame, Size sz, Position base ) {
	this(frame, new StarList(), new MapFunction(), sz, base, 10.0);
    }

    public CatalogCanvas ( Frame frame, StarList list, MapFunction mf, Size sz, Position base, double mag ) {
	parent = frame;
	catalog_list = list;
	map_function = mf.inverse();
	bg_image = null;
	size = new Size(sz.getWidth(), sz.getHeight());
	base_pos = new Position(base.getX(), base.getY());
	mark_pt = new Point(-1, -1);
	mark_color = Color.magenta;
	mag_limit = mag;
	mark_center = false;
	east_right = false;
	bg_black = false;
	frame_line = false;

	dialog_enable = true;

	draw_foreground = true;
	draw_background = true;

	resize(sz.getWidth(), sz.getHeight());
    }

    public void markCenter ( boolean flag ) {
	mark_center = flag;
    }

    public void eastRight ( boolean flag ) {
	east_right = flag;
    }

    public void bgBlack ( boolean flag ) {
	bg_black = flag;
    }
    
    public boolean isMarkCenter ( ) {
	return mark_center;
    }

    public boolean isEastRight ( ) {
	return east_right;
    }

    public boolean isBgBlack ( ) {
	return bg_black;
    }

    public void setCatalog ( StarList list, double mag ) {
	catalog_list = list;
	mag_limit = mag;
    }

    public void setCatalog ( StarList list, MapFunction mf, double mag ) {
	catalog_list = list;
	map_function = mf.inverse();
	mag_limit = mag;
    }

    public void setMark( Point pt ) {
	mark_pt = new Point(pt.x, pt.y);
    }

    public void setMarkColor( Color c ) {
	mark_color = c;
    }

    public void setImage ( Image image) {
	bg_image = image;
    }

    public void clearImage ( ) {
	bg_image = null;
    }

    public void enableDialog( ) {
	dialog_enable = true;
    }

    public void disableDialog( ) {
	dialog_enable = false;
    }

    public void enableDrawForeground ( ) {
	draw_foreground = true;
    }

    public void disableDrawForeground ( ) {
	draw_foreground = false;
    }

    public void enableDrawBackground ( ) {
	draw_background = true;
    }

    public void disableDrawBackground ( ) {
	draw_background = false;
    }

    public Color getStarColor ( ) {
	if (bg_image != null) {
	    return Color.green;
	} else {
	    if (isBgBlack())
		return Color.white;
	    else
		return Color.black;
	}
    }

    protected Point getStarPoint ( Position p ) {
	Position pos = map_function.map(p);
	Point pt = new Point((int)((isEastRight() ? -1.0 : 1.0) * pos.getX() + base_pos.getX()), (int)(pos.getY() + base_pos.getY()));
	return pt;
    }

    public boolean mouseDown ( Event e, int x, int y ) {
	if (dialog_enable  &&  parent != null) {
	    // output data of a star at the mouse-clicked point
	    Vector string_list = new Vector();

	    StarData sd = catalog_list.getFirstElement();
	    while (sd != null) {
		StringTokenizer tokens = getStarRecordTokensAt(sd, x, y);
		while (tokens != null  &&  tokens.hasMoreTokens())
			string_list.addElement(tokens.nextToken());

		StarData sd1 = sd.getFirstAppendix();
		while (sd1 != null) {
		    tokens = getStarRecordTokensAt(sd1, x, y);
		    while (tokens != null  &&  tokens.hasMoreTokens())
			string_list.addElement(tokens.nextToken());
		    sd1 = sd.getNextAppendix();
		}

		sd = catalog_list.getNextElement();
	    }

	    if (string_list.size() > 0) {
		InfoDialog dlg = new InfoDialog(parent, string_list);
		dlg.resize(size.getWidth(), size.getHeight() / 3);
		dlg.show();
	    }

	    return true;
	}
	return false;
    }

    private StringTokenizer getStarRecordTokensAt ( StarData sd, int x, int y ) {
	Point pt = getStarPoint(sd.getPosition());

	int d = (int)(mag_limit - sd.getMag()) + 1;
	if (d < 1)
	    d = 1;
	if (sd.isCircle()  &&  d < 3)
	    d = 3;

	if (Math.abs(pt.x - x) <= d  &&  Math.abs(pt.y - y) <= d) {
	    StringTokenizer tokens = new StringTokenizer(sd.getRecord(), "\n");
	    return tokens;
	}

	return null;
    }

    private Dimension offSize;
    private Graphics offScreen;
    private Image offImage;

    public void updateOffScreen ( ) {
	// clear
	if (isBgBlack())
	    offScreen.setColor(Color.black);
	else
	    offScreen.setColor(Color.white);
	offScreen.fillRect(0, 0, size.getWidth(), size.getHeight());

	// background image
	if (bg_image != null)
	    offScreen.drawImage(bg_image, 0, 0, size.getWidth(), size.getHeight(), this);

	if (draw_background)
	    drawBackground(offScreen, size);

	StarData sd = catalog_list.getFirstElement();
	while (sd != null) {
	    plotStarData(sd);

	    StarData sd1 = sd.getFirstAppendix();
	    while (sd1 != null) {
		plotStarData(sd1);
		sd1 = sd.getNextAppendix();
	    }

	    sd = catalog_list.getNextElement();
	}

	if (isMarkCenter()) {
	    offScreen.setColor(Color.red);

	    offScreen.drawArc(size.getWidth() / 2 - 7, size.getHeight() / 2 - 7, 14, 14, 0, 360);
	    offScreen.drawArc(size.getWidth() / 2 - 6, size.getHeight()  / 2 - 6, 12, 12, 0, 360);
	    offScreen.drawArc(size.getWidth() / 2 - 5, size.getHeight()  / 2 - 5, 10, 10, 0, 360);
	}

	// frameline
	if (frame_line) {
	    if (isBgBlack())
		offScreen.setColor(Color.white);
	    else
		offScreen.setColor(Color.black);
	    offScreen.drawRect(0, 0, size.getWidth(), size.getHeight());
	}

	if (draw_foreground)
	    drawForeground(offScreen, size);
    }

    protected void drawForeground ( Graphics g, Size s ) {
    }

    protected void drawBackground ( Graphics g, Size s ) {
    }

    private void plotStarData ( StarData sd ) {
	if (sd.getColor() == null)
	    offScreen.setColor(getStarColor());
	else
	    offScreen.setColor(sd.getColor());

	Point pt = getStarPoint(sd.getPosition());
	if (0 <= pt.x  &&  pt.x < size.getWidth()  &&  0 <= pt.y  &&  pt.y < size.getHeight()) {
	    if (sd.isCircle()) {
		int d = (int)(mag_limit - sd.getMag()) + 1;
		if (d < 3)
		    d = 3;

		offScreen.drawArc(pt.x - d - 1, pt.y - d - 1, 2*(d+1), 2*(d+1), 0, 360);
		offScreen.drawArc(pt.x - d, pt.y - d, 2*d, 2*d, 0, 360);
		offScreen.drawArc(pt.x - d + 1, pt.y - d + 1, 2*(d-1), 2*(d-1), 0, 360);
	    } else {
		int d = (int)(mag_limit - sd.getMag()) + 1;
		if (d < 1) {
		    offScreen.fillRect(pt.x, pt.y, 1, 1);
		} else {
		    pt.x -= d;
		    pt.y -= d;
		    offScreen.fillArc(pt.x, pt.y, d * 2, d * 2, 0, 360);
		}
	    }
	}
    }

    public void paint ( Graphics g ) {
	update(g);
    }

    public void update ( Graphics g ) {
	Dimension dim = size();
	if (offScreen == null  ||  dim.width != offSize.width ||  dim.height != offSize.height) {
	    offSize = dim;
	    offImage = createImage(dim.width, dim.height);
	    offScreen = offImage.getGraphics();

	    updateOffScreen();
	}

	g.drawImage(offImage, 0, 0, this);

	// mark
	if (0 <= mark_pt.x  &&  mark_pt.x < size.getWidth()  &&  0 <= mark_pt.y  &&  mark_pt.y < size.getHeight()) {
	    g.setColor(mark_color);
	    g.drawOval(mark_pt.x - 6, mark_pt.y - 6, 12, 12);
	    g.drawOval(mark_pt.x - 7, mark_pt.y - 7, 14, 14);
	    g.drawOval(mark_pt.x - 8, mark_pt.y - 8, 16, 16);
	}
    }
}
