/*
 * ImageRGB.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.util;

import java.awt.*;
import java.awt.image.*;

public class ImageRGB {
    private ImageGray RImage;
    private ImageGray GImage;
    private ImageGray BImage;

    public ImageRGB ( ) {
	RImage = new ImageGray();
	GImage = new ImageGray();
	BImage = new ImageGray();
    }

    public ImageRGB ( Size size ) throws Exception {
	allocate(size);
    }

    public ImageRGB ( ImageRGB rgb_image ) {
	RImage = new ImageGray(rgb_image.RImage);
	GImage = new ImageGray(rgb_image.GImage);
	BImage = new ImageGray(rgb_image.BImage);
    }

    public ImageRGB ( Image image, Size size ) throws Exception {
	// store the image into an array
	int ipixels[] = new int[size.getWidth() * size.getHeight()];
	PixelGrabber pg = new PixelGrabber(image, 0, 0, size.getWidth(), size.getHeight(), ipixels, 0, size.getWidth());
	try {
	    pg.grabPixels();
	} catch (InterruptedException e) {
	    System.err.println("Interrupted waiting for pixels.");
	    throw e;
	}
	if ((pg.status() & ImageObserver.ABORT) != 0) {
	    System.err.println("Image fetch aborted or errored.");
	    throw new Exception();
	}

	try {
	    allocate(size);
	} catch ( Exception e ) {
	    throw e;
	}

	Point pt = new Point(0, 0);
	int i = 0;
	for (pt.y = 0 ; pt.y < size.getHeight() ; pt.y++) {
	    for (pt.x = 0 ; pt.x < size.getWidth() ; pt.x++) {
		setRValue(pt, (double)((ipixels[i] & (255 << 16)) >> 16));
		setGValue(pt, (double)((ipixels[i] & (255 << 8)) >> 8));
		setBValue(pt, (double)(ipixels[i] & 255));
		i++;
	    }
	}
    }

    protected void allocate ( Size size ) throws Exception {
	try {
	    RImage = new ImageGray(size);
	    GImage = new ImageGray(size);
	    BImage = new ImageGray(size);
	} catch (Exception e) {
	    RImage = new ImageGray();
	    GImage = new ImageGray();
	    BImage = new ImageGray();
	    throw e;
	}
    }

    public Size getSize ( ) {
	return RImage.getSize();
    }

    public int getWidth ( ) {
	return RImage.getWidth();
    }

    public int getHeight ( ) {
	return RImage.getHeight();
    }

    public void setRValue ( Point pt, double val ) {
	RImage.setValue(pt, val);
    }

    public void setGValue ( Point pt, double val ) {
	GImage.setValue(pt, val);
    }

    public void setBValue ( Point pt, double val ) {
	BImage.setValue(pt, val);
    }

    public double getRValue ( Point pt ) {
	return RImage.getValue(pt);
    }

    public double getGValue ( Point pt ) {
	return GImage.getValue(pt);
    }

    public double getBValue ( Point pt ) {
	return BImage.getValue(pt);
    }

    public double getGrayValue ( Point pt ) {
	return 0.299 * getRValue(pt) + 0.587 * getGValue(pt) + 0.114 * getBValue(pt);
    }

    public MinMaxAverage calculateMinMaxAverage ( ) {
	MinMaxAverage mmaR = RImage.calculateMinMaxAverage();
	MinMaxAverage mmaG = GImage.calculateMinMaxAverage();
	MinMaxAverage mmaB = BImage.calculateMinMaxAverage();

	double min = mmaR.getMin();
	double max = mmaR.getMax();
	if (min > mmaG.getMin())
	    min = mmaG.getMin();
	if (max < mmaG.getMax())
	    max = mmaG.getMax();
	if (min > mmaB.getMin())
	    min = mmaB.getMin();
	if (max < mmaB.getMax())
	    max = mmaB.getMax();

	return new MinMaxAverage(min, max, (mmaR.getAverage() + mmaG.getAverage() + mmaB.getAverage()) / 3.0, mmaR.getCount() + mmaG.getCount() + mmaB.getCount());
    }

    public ImageGray getGrayImage ( ) throws Exception {
	ImageGray image;
	try {
	    image = new ImageGray(getSize());
	} catch ( Exception e ) {
	    throw e;
	}

	Point pt = new Point(0, 0);
	for (pt.y = 0 ; pt.y < getHeight() ; pt.y++) {
	    for (pt.x = 0 ; pt.x < getWidth() ; pt.x++) {
		image.setValue(pt, getGrayValue(pt));
	    }
	}

	return image;
    }

    public Image getImage ( ) throws Exception {
	MinMaxAverage mma = calculateMinMaxAverage();
	try {
	    return getImage(mma.getMin(), mma.getMax());
	} catch (Exception e) {
	    throw e;
	}
    }

    public Image getImage ( double min, double max ) throws Exception {
	int p[] = new int[getWidth() * getHeight()];
	if (p == null) {
	    System.err.println("Out of memory. Cannot allocate Image.");
	    throw new Exception();
	}

	Point pt = new Point(0, 0);
	int r, g, b;
	int i = 0;
	for (pt.y = 0 ; pt.y < getHeight() ; pt.y++) {
	    for (pt.x = 0 ; pt.x < getWidth() ; pt.x++) {
		r = (int)((getRValue(pt) - min) / (max - min) * 255);
		if (r < 0)
		    r = 0;
		if (r > 255)
		    r = 255;
		g = (int)((getGValue(pt) - min) / (max - min) * 255);
		if (g < 0)
		    g = 0;
		if (g > 255)
		    g = 255;
		b = (int)((getBValue(pt) - min) / (max - min) * 255);
		if (b < 0)
		    b = 0;
		if (b > 255)
		    b = 255;
		p[i] = (255 << 24) | (r << 16) | (g << 8) | b;
		i++;
	    }
	}
	Image image = Toolkit.getDefaultToolkit().createImage(new MemoryImageSource(getWidth(), getHeight(), p, 0, getWidth()));
	return image;
    }

    public ImageRGB resize ( Size new_size ) throws Exception {
	try {
	    ImageRGB rgb_image = null;
	    rgb_image.RImage = RImage.resize(new_size);
	    rgb_image.GImage = GImage.resize(new_size);
	    rgb_image.BImage = BImage.resize(new_size);
	    return rgb_image;
	} catch ( Exception e ) {
	    throw e;
	}
    }

    // expansion and reduction
    public ImageRGB rescale ( Size new_size ) throws Exception {
	try {
	    ImageRGB rgb_image = null;
	    rgb_image.RImage = RImage.rescale(new_size);
	    rgb_image.GImage = GImage.rescale(new_size);
	    rgb_image.BImage = BImage.rescale(new_size);
	    return rgb_image;
	} catch ( Exception e ) {
	    throw e;
	}
    }

    public ImageRGB rotate ( double theta ) throws Exception {
	try {
	    ImageRGB rgb_image = null;
	    rgb_image.RImage = RImage.rotate(theta);
	    rgb_image.GImage = GImage.rotate(theta);
	    rgb_image.BImage = BImage.rotate(theta);
	    return rgb_image;
	} catch ( Exception e ) {
	    throw e;
	}
    }

    // left <-> right
    public void mirrorHorizontal ( ) {
	RImage.mirrorHorizontal();
	GImage.mirrorHorizontal();
	BImage.mirrorHorizontal();
    }

    // upside down
    public void mirrorVertical ( ) {
	RImage.mirrorVertical();
	GImage.mirrorVertical();
	BImage.mirrorVertical();
    }
}
